// ExecutionEvents.h
// Place holder for the interceiptable FLIP execution events,
// encompassing all necessary callbacks, providing the
// typical "set" and "notification" functionality.
// A. Savidis, deep re-factoring of first version from Y. Lilis.
// September, 2005.
// Project: FLIP Suite (Execution visualizer, Savidis, 2005).
//

#ifdef	HAS_VISUALIZER

#ifndef	EXECUTIONEVENTS_H

#include <stdio.h>
#include <string>
using namespace std;

///////////////////////////////////////////////////////////////////

class ExecutionEvents {

	private:

	static void (*onBlockEnter)(unsigned line);
	static void (*onBlockExit)(unsigned line);
	static void (*onIfEnter)(unsigned line);
	static void (*onIfExit)(unsigned line);
	static void (*onForEnter)(unsigned line);
	static void (*onForExit)(unsigned line);
	static void (*onWhileEnter)(unsigned line);
	static void (*onWhileExit)(unsigned line);
	static void (*onBreak)(unsigned line);
	static void (*onCaseBegin)(unsigned line);
	static void (*onCaseEnd)(unsigned line);
	static void (*onContinue)(unsigned line);
	static void (*onConditionValue)(const char* stmt, bool val);

	//*************************

	static void (*onArithmeticOperatorUsage)(
					const char*		op, 
					unsigned		page1, // arg1
					unsigned		row1, 
					unsigned		page2, // arg2
					unsigned		row2
				);

	static void (*onRelationalOperatorUsage)(
					const char*		op, 
					unsigned		page1, // arg1
					unsigned		row1, 
					unsigned		page2, // arg2
					unsigned		row2
				);

	static void (*onUnaryOperatorUsage)(
					const char*		op, 
					unsigned		page, 
					unsigned		row
				);

	//*************************

	static void (*onVarWrite)(
					const std::string&	value, 
					const char*			type,	// The FLIP data type.
					unsigned			page, 
					unsigned			row
				);

	static void (*onVarDeclaration)(
					const std::string&	name, 
					const std::string&	value, 
					const char*			type,		// The FLIP data type.
					const char*			storage,	// The FLIP storage class.
					unsigned			page, 
					unsigned			row
				);

	static void (*onVarDestruction)(unsigned page, unsigned row);

	//*************************

	static void (*onArrayDeclaration)(
					unsigned startPage, 
					unsigned startRow, 
					unsigned length
				);

	static void (*onArrayAllocation)(
					unsigned length
				);

	static void (*onArrayDestruction)(
					unsigned startPage, 
					unsigned startRow, 
					unsigned length
				);

	static void (*onArrayIndexing)(
					unsigned	startPage,
					unsigned	startRow,
					unsigned	length,
					unsigned	index
				);

	//*************************

	static void (*onRecordDeclaration)(void);
	static void (*onRecordDestruction)(void);

	//*************************

	static void (*onFunctionEnter)(const char* funname, unsigned line);
	static void (*onFunctionExit)(const char* funname, unsigned line);
	static void (*onFunctionUsage)(const char *type, const char *name, unsigned line);

	static void (*onResult)(const char* value, const char* funname, unsigned line);
	static void (*onFinish)(const char* funname, unsigned line);

	//*************************

	static void (*onPageOf)(unsigned page, unsigned row, unsigned line);
	static void (*onRowOf)(unsigned page, unsigned row, unsigned line);
	static void (*onContentOf)(unsigned page, unsigned row, unsigned line);
	static void (*onPositionOf)(unsigned page, unsigned row, unsigned line);
	
	static void (*onGrantPositions)(
						unsigned page, 
						unsigned row, 
						unsigned size, 
						unsigned line
					);

	static void (*onReleasePositions)(
						unsigned page, 
						unsigned row, 
						unsigned line
					);

	//*************************

	static void (*onRegionUsage)(
						unsigned page, 
						unsigned startRow, 
						unsigned endRow, 
						unsigned line
					);

	static void (*onRegionIndexing)(
						unsigned page, 
						unsigned startRow, 
						unsigned endRow, 
						unsigned index
					);
	
	//*************************

	static void			(*onProgramStart)(void);
	static void			(*onProgramEnd)(void);
	static void			(*onProgramError)(const char* s);
	static void			(*onProgramStop)(unsigned line);

	static int			(*outputFun)(const char *format, ...);
	static const char*	(*inputFun)(const char* buffer, unsigned len, FILE* fp);
	static void			(*cleanUpFun)(void);

	///////////////////////////////////////////////////////////////////

	public:

	static void SetOnBlockEnter (void (*f)(unsigned line))
					{  onBlockEnter = f; }

	static void SetOnBlockExit (void (*f)(unsigned line))
					{ onBlockExit = f; }

	static void SetOnIfEnter (void (*f)(unsigned line))
					{ onIfEnter = f; }

	static void SetOnIfExit (void (*f)(unsigned line))
					{ onIfExit = f; }

	static void SetOnForEnter (void (*f)(unsigned line))
					{ onForEnter = f; }

	static void SetOnForExit (void (*f)(unsigned line))
					{ onForExit = f; }

	static void SetOnWhileEnter (void (*f)(unsigned line))
					{ onWhileEnter = f; }

	static void SetOnWhileExit (void (*f)(unsigned line))
					{ onWhileExit = f; }

	static void SetOnBreak (void (*f)(unsigned line))
					{ onBreak = f; }

	static void SetOnCaseBegin (void (*f)(unsigned line))
					{ onCaseBegin = f; }

	static void SetOnCaseEnd (void (*f)(unsigned line))
					{ onCaseEnd = f; }

	static void SetOnContinue (void (*f)(unsigned line))
					{ onContinue = f; }

	static void SetOnConditionValue (void (*f)(const char* s, bool v))
					{ onConditionValue = f; }

	static void NotifyBlockEnter (unsigned line)
					{ (*onBlockEnter)(line); }

	static void NotifyBlockExit (unsigned line)
					{ (*onBlockExit)(line); }

	static void NotifyIfEnter (unsigned line)
					{ (*onIfEnter)(line); }

	static void NotifyIfExit (unsigned line)
					{ (*onIfExit)(line); }

	static void NotifyForEnter (unsigned line)
					{ (*onForEnter)(line); }

	static void NotifyForExit (unsigned line)
					{ (*onForExit)(line); }

	static void NotifyWhileEnter (unsigned line)
					{ (*onWhileEnter)(line); }

	static void NotifyWhileExit (unsigned line)
					{ (*onWhileExit)(line); }

	static void NotifyBreak (unsigned line)
					{ (*onBreak)(line); }

	static void NotifyContinue (unsigned line)
					{ (*onContinue)(line); }

	static void NotifyCaseBegin (unsigned line)
					{ (*onCaseBegin)(line); }

	static void NotifyCaseEnd (unsigned line)
					{ (*onCaseEnd)(line); }

	static void NotifyConditionValue(
					const char* stmt, 
					bool		val
				) { (*onConditionValue)(stmt, val); }

#define	NOTIFY_BLOCKENTER(l)				ExecutionEvents::NotifyBlockEnter(l)
#define	NOTIFY_BLOCKEXIT(l)					ExecutionEvents::NotifyBlockExit(l)
#define	NOTIFY_IFENTER(l)					ExecutionEvents::NotifyIfEnter(l)
#define	NOTIFY_IFEXIT(l)					ExecutionEvents::NotifyIfExit(l)
#define	NOTIFY_FORENTER(l)					ExecutionEvents::NotifyForEnter(l)
#define	NOTIFY_FOREXIT(l)					ExecutionEvents::NotifyForExit(l)
#define	NOTIFY_WHILEENTER(l)				ExecutionEvents::NotifyWhileEnter(l)
#define	NOTIFY_WHILEEXIT(l)					ExecutionEvents::NotifyWhileExit(l)
#define	NOTIFY_BREAK(l)						ExecutionEvents::NotifyBreak(l)
#define	NOTIFY_CONTINUE(l)					ExecutionEvents::NotifyContinue(l)
#define	NOTIFY_CASEBEGIN(l)					ExecutionEvents::NotifyCaseBegin(l)
#define	NOTIFY_CASEEND(l)					ExecutionEvents::NotifyCaseEnd(l)
#define	NOTIFY_CONDITIONVALUE(stmt, val)	ExecutionEvents::NotifyConditionValue(stmt, val)

	//*************************
	// OPERATORS

	static void SetOnArithmeticOperatorUsage(
					void (*f)(
						const char* op, 
						unsigned	page1, 
						unsigned	row1, 
						unsigned	page2, 
						unsigned	row2
					)
				){ onArithmeticOperatorUsage = f; }

	static void SetOnRelationalOperatorUsage(
					void (*f)(
						const char* op, 
						unsigned	page1, 
						unsigned	row1, 
						unsigned	page2, 
						unsigned	row2
					)
				){ onRelationalOperatorUsage = f; }

	static void SetOnUnaryOperatorUsage(
					void (*f)(
						const char*	op, 
						unsigned	page, 
						unsigned	row
					)
				){ onUnaryOperatorUsage = f; }

	static void	NotifyArithmeticOperatorUsage (
					const char* op, 
					unsigned	page1, 
					unsigned	row1, 
					unsigned	page2, 
					unsigned	row2
					) { (*onArithmeticOperatorUsage)(op, page1, row1, page2, row2); }

	static void	NotifyRelationalOperatorUsage (
					const char* op, 
					unsigned	page1, 
					unsigned	row1, 
					unsigned	page2, 
					unsigned	row2
					) { (*onRelationalOperatorUsage)(op, page1, row1, page2, row2); }

	static void	NotifyUnaryOperatorUsage (
					const char* op, 
					unsigned	page, 
					unsigned	row
					) { (*onUnaryOperatorUsage)(op, page, row); }

#define	NOTIFY_ARITHOP(op,p1,r1,p2,r2) \
	ExecutionEvents::NotifyArithmeticOperatorUsage(op,p1,r1,p2,r2)

#define	NOTIFY_RELATOP(op,p1,r1,p2,r2) \
	ExecutionEvents::NotifyRelationalOperatorUsage(op,p1,r1,p2,r2)

#define	NOTIFY_UNARYOP(op,p,r) \
	ExecutionEvents::NotifyUnaryOperatorUsage(op,p,r)

	//*************************
	// VARIABLES

	static void SetOnVarWrite(
					void (*f)(
						const std::string&	value, 
						const char*			type,
						unsigned			page, 
						unsigned			row
					)
				){ onVarWrite = f; }

	static void SetOnVarDeclaration(
					void (*f) (
						const std::string&	name, 
						const std::string&	value, 
						const char*			type,
						const char*			storage,
						unsigned			page, 
						unsigned			row
					)
				){ onVarDeclaration = f; }

	static void SetOnVarDestruction(
					void (*f)(unsigned page, unsigned row)
					){ onVarDestruction = f; }

	static void NotifyVarWrite (
					const std::string&	value, 
					const char*			type,
					unsigned			page, 
					unsigned			row
				) { (*onVarWrite)(value, type, page, row); }

	static void NotifyVarDeclaration (
						const std::string&	name, 
						const std::string&	value, 
						const char*			type,
						const char*			storage,
						unsigned			page, 
						unsigned			row
				) { (*onVarDeclaration)(name, value, type, storage, page, row); }

	static void NotifyVarDestruction (
						unsigned		page, 
						unsigned		row
				) { (*onVarDestruction)(page, row); }

#define	NOTIFY_VARWRITE(v,t,p,r)	ExecutionEvents::NotifyVarWrite(v,t,p,r)
#define	NOTIFY_VARDECL(n,v,t,s,p,r)	ExecutionEvents::NotifyVarDeclaration(n,v,t,s,p,r)
#define	NOTIFY_VARDESTR(p,r)		ExecutionEvents::NotifyVarDestruction(p,r)

	//*************************
	// ARRAYS

	static void SetOnArrayDeclaration(
					void(*f)(
						unsigned startPage, 
						unsigned startRow, 
						unsigned length
					)
				){ onArrayDeclaration = f; }

	static void SetOnArrayAllocation(
					void(*f)(
						unsigned length
					)
				){ onArrayAllocation = f; }

	static void SetOnArrayDestruction(
					void(*f)(
						unsigned startPage, 
						unsigned startRow, 
						unsigned length
					)
				) { onArrayDestruction = f; }

	static void SetOnArrayIndexing(
					void(*f)(
						unsigned	startPage, 
						unsigned	startRow, 
						unsigned	length, 
						unsigned	index
					)
				) { onArrayIndexing = f; }

	static void NotifyArrayDeclaration (
					unsigned startPage, 
					unsigned startRow, 
					unsigned length
				) { (*onArrayDeclaration)(startPage, startRow, length); }

	static void NotifyArrayAllocation (
					unsigned length
				) { (*onArrayAllocation)(length); }

	static void NotifyArrayDestruction (
					unsigned startPage, 
					unsigned startRow, 
					unsigned line
				) { (*onArrayDestruction)(startPage, startRow, line); }

	static void NotifyArrayIndexing (
					unsigned startPage, 
					unsigned startRow, 
					unsigned length,
					unsigned index
				) { (*onArrayIndexing)(startPage, startRow, length, index); }

#define	NOTIFY_ARRAYDECL(sp,sr,len)		ExecutionEvents::NotifyArrayDeclaration(sp,sr,len)
#define	NOTIFY_ARRAYALLOC(len)			ExecutionEvents::NotifyArrayAllocation(len)
#define	NOTIFY_ARRAYDESTR(sp,sr,l)		ExecutionEvents::NotifyArrayDestruction(sp,sr,l)
#define	NOTIFY_ARRAYINDEX(sp,sr,len,i)	ExecutionEvents::NotifyArrayIndexing(sp,sr,len,i)

	//*************************
	// RECORDS

	static void SetOnRecordDeclaration (void(*f)(void))
				{ onRecordDeclaration = f; }

	static void SetOnRecordDestruction (void(*f)(void))
				{ onRecordDestruction = f; }

	static void NotifyRecordDeclaration (void)
				{ (*onRecordDeclaration)(); }

	static void NotifyRecordDestruction (void)
				{ (*onRecordDestruction)(); }

#define	NOTIFY_RECORDDECL()		ExecutionEvents::NotifyRecordDeclaration()
#define	NOTIFY_RECORDDESTR()	ExecutionEvents::NotifyRecordDestruction()

	//*************************
	// FUNCTIONS

	static void SetOnFunctionEnter (void (*f)(const char* name, unsigned line))
		{ onFunctionEnter = f; }

	static void SetOnFunctionExit (void (*f)(const char* name, unsigned line))
		{ onFunctionExit = f; }

	static void SetOnFunctionUsage (
					void (*f)(
							const char* type, 
							const char*	name, 
							unsigned	line
					)
				) { onFunctionUsage = f; }

	static void SetOnResult(
					void (*f)(
						const char*	value, 
						const char*	name, 
						unsigned	line
					)
				) { onResult = f; }

	static void SetOnFinish(
					void (*f)(
						const char*	name, 
						unsigned	line
					)
				){ onFinish = f; }
	
	static void NotifyFunctionEnter (const char* name, unsigned line)
		{ (*onFunctionEnter)(name, line); }

	static void NotifyFunctionExit (const char* name, unsigned line)
		{ (*onFunctionExit)(name, line); }

	static void NotifyFunctionUsage (const char* type, const char* name, unsigned line)
		{ (*onFunctionUsage)(type, name, line); }

	// Return type is necessary merely because of the way the notification
	// function is injected in servant.
	//
	static bool NotifyResult (const char* value, const char* name, unsigned line)
		{ (*onResult)(value, name, line); return true; }

	static void NotifyFinish (const char* name, unsigned line)
		{ (*onFinish)(name, line); }

#define	NOTIFY_FUNCENTER(n,l)	ExecutionEvents::NotifyFunctionEnter(n,l)
#define	NOTIFY_FUNCEXIT(n,l)	ExecutionEvents::NotifyFunctionExit(n,l)
#define	NOTIFY_FUNCUSE(t,n,l)	ExecutionEvents::NotifyFunctionUsage(t,n,l)
#define	NOTIFY_RESULT(v,n,l)	ExecutionEvents::NotifyResult(v,n,l)
#define	NOTIFY_FINISH(n,l)		ExecutionEvents::NotifyFinish(n,l)

	//*************************
	// MEMORY USE AND ALLOCATION.
	//
	static void SetOnPageOf (void (*f)(unsigned page, unsigned row, unsigned line))
		{ onPageOf = f; }

	static void SetOnRowOf (void (*f)(unsigned page, unsigned row, unsigned line))
		{ onRowOf = f; }

	static void SetOnContentOf (void (*f)(unsigned page, unsigned row, unsigned line))
		{ onContentOf = f; }

	static void SetOnPositionOf (void (*f)(unsigned page, unsigned row, unsigned line))
		{ onPositionOf = f; }
	
	static void SetOnGrantPositions(
					void (*f) (
						unsigned page, 
						unsigned row, 
						unsigned size, 
						unsigned line
					)
				) { onGrantPositions = f; }

	static void SetOnReleasePositions (
					void (*f) (
						unsigned page, 
						unsigned row, 
						unsigned line
					)
				) { onReleasePositions = f; }

	static void NotifyPageOf (unsigned page, unsigned row, unsigned line)
		{ (*onPageOf)(page, row, line); }

	static void NotifyRowOf (unsigned page, unsigned row, unsigned line)
		{ (*onRowOf)(page, row, line); }

	static void NotifyContentOf (unsigned page, unsigned row, unsigned line)
		{ (*onContentOf)(page, row, line); }

	static void NotifyPositionOf (unsigned page, unsigned row, unsigned line)
		{ (*onPositionOf)(page, row, line); }

	static void NotifyGrantPositions (
					unsigned page, 
					unsigned row, 
					unsigned size, 
					unsigned line
				) { (*onGrantPositions)(page, row, size, line); }

	static void NotifyReleasePositions (
					unsigned page, 
					unsigned row, 
					unsigned line
				) { (*onReleasePositions)(page, row, line); }

#define	NOTIFY_PAGEOF(p,r,l)		ExecutionEvents::NotifyPageOf(p,r,l)
#define	NOTIFY_ROWOF(p,r,l)			ExecutionEvents::NotifyRowOf(p,r,l)
#define	NOTIFY_CONTENTOF(p,r,l)		ExecutionEvents::NotifyContentOf(p,r,l)
#define	NOTIFY_POSITIONOF(p,r,l)	ExecutionEvents::NotifyPositionOf(p,r,l)
#define	NOTIFY_GRANT(p,r,s,l)		ExecutionEvents::NotifyGrantPositions(p,r,s,l)
#define	NOTIFY_RELEASE(p,r,l)		ExecutionEvents::NotifyReleasePositions(p,r,l)

	//*************************
	// REGIONS.
	//
	static void SetOnRegionUsage(
					void (*f) (
						unsigned page, 
						unsigned startRow, 
						unsigned endRow, 
						unsigned line
					)
				) { onRegionUsage = f; }

	static void SetOnRegionIndexing (
					void (*f) (
						unsigned page, 
						unsigned startRow, 
						unsigned endRow, 
						unsigned index
					)
				) { onRegionIndexing = f; }

	static void NotifyRegionUsage(
					unsigned page, 
					unsigned startRow, 
					unsigned endRow, 
					unsigned line
				) { (*onRegionUsage)(page, startRow, endRow, line); }

	static void NotifyRegionIndexing(
					unsigned page, 
					unsigned startRow, 
					unsigned endRow, 
					unsigned index
				) { (*onRegionIndexing)(page, startRow, endRow, index); }

#define	NOTIFY_REGIONUSE(p,sr,er,l)		ExecutionEvents::NotifyRegionUsage(p,sr,er,l)
#define	NOTIFY_REGIONINDEX(p,sr,er,i)	ExecutionEvents::NotifyRegionIndexing(p,sr,er,i)

	//*************************
	// PRORGAM RUN AND I/O.
	
	static void SetOnProgramStart (void (*f)(void)) { onProgramStart = f; }
	static void SetOnProgramEnd (void (*f)(void)) { onProgramEnd = f; }
	static void SetOnProgramError (void (*f)(const char* s)) { onProgramError = f; }
	static void SetOnProgramStop (void (*f)(unsigned line)) { onProgramStop = f; }

	static void NotifyProgramStart (void) { (*onProgramStart)(); }
	static void NotifyProgramEnd (void) { (*onProgramEnd)(); }
	static void NotifyProgramError (const char* s) { (*onProgramError)(s); }
	static void NotifyProgramStop (unsigned l) { (*onProgramStop)(l); }

#define	NOTIFY_PROGRAMSTART()		ExecutionEvents::NotifyProgramStart()
#define	NOTIFY_PROGRAMEND()			ExecutionEvents::NotifyProgramEnd()
#define	NOTIFY_PROGRAMERROR(s)		ExecutionEvents::NotifyProgramError(s)
#define	NOTIFY_PROGRAMSTOP(l)		ExecutionEvents::NotifyProgramStop(l)

	static void SetOutputFunction (int (*f)(const char *format, ...))
		{ outputFun = f; }

	static void SetInputFunction (const char* (*f)(const char *buf, unsigned len, FILE* fp))
		{ inputFun = f; }

	static void SetCleanUpFunction (void (*f)(void))
		{ cleanUpFun = f; }

	typedef int (*OutputFuncType)(const char *format, ...);
	typedef const char* (*InputFuncType)(const char *buff, unsigned len, FILE* fp);
	typedef void (*CleanUpFuncType) (void);

	static OutputFuncType GetOutputFunction (void) 
		{ return outputFun; }

	static InputFuncType GetInputFunction (void) 
		{ return inputFun; }

	static CleanUpFuncType GetCleanUpFunction (void) 
		{ return cleanUpFun; }

#define	INPUT_FUNCTION		ExecutionEvents::GetInputFunction()
#define	OUTPUT_FUNCTION		ExecutionEvents::GetOutputFunction()
#define	CLEANUP_FUNCTION	ExecutionEvents::GetCleanUpFunction()

};

#endif	// Dop not add stuff beyond this point.

#endif	// HAS_VISUALIZER
