///////////////////////////////////////////////////////
// Servant.h
// Anthony Savidis.
// Version 2.0/2007 
// Apart from language extensions, V2.0 includes the
// linkage to FLIP visualisers.
///////////////////////////////////////////////////////
// FLIP language bindings implemented for the C++ language.
// FLIP is an imperative language for the Computational
// Servant whose name is formed by the initials: 
// Front-end Language for Introduction to Programming.
///////////////////////////////////////////////////////


#ifndef	SERVANT_H
#define	SERVANT_H

#pragma warning(disable:4786)
#pragma warning(disable:4996)

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <time.h>

#include <string>
#include <map>
#include <algorithm>
#include <functional>
#include <utility>
#include <list>
using namespace std;

unsigned static ProgramStartLine (void);

///////////////////////////////////////////////////////

#if	defined(HAS_VISUALIZER) && (defined(CONSOLE_VISUALIZER) || defined(GRAPHICAL_VISUALIZER))
#include "VisualiserSuper.h"
#else

// If we do not have a visualizer we simply strip-off the
// notification calls.
//
#define	NOTIFY_BLOCKENTER(l)
#define	NOTIFY_BLOCKEXIT(l)
#define	NOTIFY_IFENTER(l)
#define	NOTIFY_IFEXIT(l)
#define	NOTIFY_FORENTER(l)
#define	NOTIFY_FOREXIT(l)
#define	NOTIFY_WHILEENTER(l)
#define	NOTIFY_WHILEEXIT(l)
#define	NOTIFY_ARITHOP(op,p1,r1,p2,r2)
#define	NOTIFY_RELATOP(op,p1,r1,p2,r2)
#define	NOTIFY_UNARYOP(op,p,r)
#define	NOTIFY_VARWRITE(v,t,p,r)
#define	NOTIFY_VARDECL(n,v,t,s,p,r)
#define	NOTIFY_VARDESTR(p,r)
#define	NOTIFY_ARRAYDECL(sp,sr,len)
#define	NOTIFY_ARRAYALLOC(len)
#define	NOTIFY_ARRAYDESTR(sp,sr,l)
#define	NOTIFY_ARRAYINDEX(sp,sr,len,i)
#define	NOTIFY_RECORDDECL()
#define	NOTIFY_RECORDDESTR()
#define	NOTIFY_FUNCENTER(n,l)
#define	NOTIFY_FUNCEXIT(n,l)
#define	NOTIFY_FUNCUSE(t,n,l)
#define	NOTIFY_RESULT(v,n,l)	true
#define	NOTIFY_FINISH(n,l)
#define	NOTIFY_PAGEOF(p,r,l)
#define	NOTIFY_ROWOF(p,r,l)
#define	NOTIFY_CONTENTOF(p,r,l)
#define	NOTIFY_POSITIONOF(p,r,l)
#define	NOTIFY_GRANT(p,r,s,l)
#define	NOTIFY_RELEASE(p,r,l)
#define	NOTIFY_REGIONUSE(p,sr,er,l)
#define	NOTIFY_REGIONINDEX(p,sr,er,i)
#define	NOTIFY_CONDITIONVALUE(stmt, val)
#define	NOTIFY_BREAK(l)
#define	NOTIFY_CONTINUE(l)
#define	NOTIFY_CASEBEGIN(l)
#define	NOTIFY_CASEEND(l)

#define	NOTIFY_PROGRAMSTART() \
		OUTPUT_FUNCTION("Servant starts running the program.\n")

#define	NOTIFY_PROGRAMEND()	\
	OUTPUT_FUNCTION("Servant finished the program normally.\n");

#define	NOTIFY_PROGRAMERROR(msg) \
	OUTPUT_FUNCTION("Error: %s\nServant has interrupted program execution.\n", msg);

#define	NOTIFY_PROGRAMSTOP(line) \
	OUTPUT_FUNCTION("Servant stopped execution at line %d.\n", line);

// Default std I/O functions and clean-up.
//
#define	INPUT_FUNCTION	fgets
#define	OUTPUT_FUNCTION	printf
inline	void CLEANUP_FUNCTION (void) { exit(1); }

#endif	// HAS_VISUALIZER

///////////////////////////////////////////////////////////////////////
// Classes for making temporary instance trapping specific statements.
//
class IfTrapper {
	unsigned line;
	public:
	IfTrapper (unsigned _line) : line(_line) { NOTIFY_IFENTER(line); }
	~IfTrapper(){ NOTIFY_IFEXIT(line); }
};

class WhileTrapper {
	unsigned line;
	public:
	WhileTrapper (unsigned _line) : line(_line) { NOTIFY_WHILEENTER(line); }
	~WhileTrapper(){ NOTIFY_WHILEEXIT(line); }
};

class ForTrapper {
	unsigned line;
	public:
	ForTrapper(unsigned _line) : line(_line) { NOTIFY_FORENTER(line);}
	~ForTrapper(){ NOTIFY_FOREXIT(line); }
};

//-------------------------------------------------------------------
// Special purpose classes for name book-keeping and recording of current line.
// Used in content-of, position and region declarations.
//
class LineNameRecorder {

	private:
	static unsigned		line;
	static string		name;

	public:
	LineNameRecorder (unsigned l, const char* n)
		{ line = l; name  = n; }

	// Reset upon use.
	//
	static const char* GetName (void) {
		assert(name != "");
		const char *s = name.c_str();
		name = "";
		return s;
	}

	static unsigned GetLine(void) {
		assert(line);
		unsigned l = line;
		line = 0;
		return l;
	}

	static bool NotConsumed (void) 
		{ return line!=0 && name != ""; }
};

string		LineNameRecorder::name = "";
unsigned	LineNameRecorder::line = 0;

/////////////////////////////////////////////////////////


#define	INSTANTIATION_TRAPPER(_class, call)		\
class _class##PreInstantiation {				\
			public:								\
			_class##PreInstantiation(void) {	\
				call;							\
			}									\
		};										\
		_class##PreInstantiation _instantiation;


#define	DESTRUCTION_TRAPPER(_class, call)		\
class _class##PreDestruction {					\
			public:								\
			~_class##PreDestruction() {			\
				call;							\
			}									\
		};										\
		_class##PreDestruction _destructrion;

/////////////////////////////////////////////////////////

class LineTrapper {
	protected:
	unsigned line;

	public:
	LineTrapper(unsigned _line) :line(_line){}
	virtual ~LineTrapper(){}
};

/////////////////////////////////////////////////////////

class Object;

Object& _CONTENTOF(const Object& x, unsigned line);
Object& _CONTENTOF(unsigned page, unsigned row, unsigned line);

class ContentOf : public LineTrapper {
	public:
	Object& operator()(const Object& x) const
			{ return(_CONTENTOF(x, line)); }

	Object& operator()(unsigned page, unsigned row) const
			{ return(_CONTENTOF(page, row, line)); }

	ContentOf (unsigned _line) : LineTrapper(_line){}
};

//-------------------------------------------------------------------
// Non-static functions are externalised (can be called from other files).
//
static void StringToUpper (char* s) {
	while (*s) {
		*s = toupper(*s);
		++s;
	}
}

//**************************

static char* SkipSpaces (char* s) {
	while (*s && isspace(*s))
		++s;
	return s;
}

//**************************

void ReadInputString (char* buff, unsigned len) {
	do{}
	while (!INPUT_FUNCTION(buff, len-1, stdin) || !*SkipSpaces(buff));
	
	// Strip-off trailing new line character.
	//
	unsigned i = strlen(buff);	
	if (buff[i-1] == '\n')
		buff[i-1] = '\0';
}

//**************************
// Read only the part of an integer number, and stop with the
// first non digit character, including null terminator.
//
static char* ReadInteger (char* s, int* x, int* signkeep = (int*) 0) {
	
	int sign =	*s == '-' ? (++s, -1)	: 
				*s == '+' ? (++s, 1)	: 1;
	if (signkeep)
		*signkeep = sign;

	for (*x = 0; isdigit(*s); ++s)
		*x = *x * 10 + *s - '0';
	*x *= sign;
	return s;
}

//**************************
// Read a pure real number with an optional fractional part.
//
static char* ReadReal (char* s, float* x) {
	
	*x = 0;
	int r, sign;
	s = ReadInteger(s, &r, &sign);

	if (*s && !isspace(*s) && *s == '.') {
		float f = (float) 0.1;
		for (++s; isdigit(*s); *x += f * (*s-'0'), f *= (float) 0.1, ++s)
			;
	}
	
	if (*(s = SkipSpaces(s)))
		{ return (char*) 0; *x = 0; }
	else {
		*x = *x * sign + r;
		return s;
	}
}

//**************************

char* ReadInputString (void) {
	static char buff[1024];
	ReadInputString(buff, 1024);
	return buff;
}

//**************************

const string MakeString (const char* format,...) {
	va_list args;
	va_start(args, format);
	static char constructStr[1024]; 
	vsprintf(constructStr, format, args);
	va_end(args);
	return constructStr;
}

///////////////////////////////////////////////////////
// Visualisers may need the command line
// arguments. If initialisation takes place before
// main, this is useless (finds 0, null).
//
static int		_servant_argc;
static char**	_servant_argv;

int		GetARGC (void) { return _servant_argc; }
char**	GetARGV (void) { return _servant_argv; }

///////////////////////////////////////////////////////
// Macros for the basic types.
//
#define INTEGER			IntObject
#define REAL			RealObject
#define STRING			StringObject
#define	BOOLEAN			BoolObject
#define	CHARACTER		CharObject
#define	ADDRESS			AddressObject

#define	TRUESTR			"TRUE"
#define	FALSESTR		"FALSE"
#define	TRUE			true
#define	FALSE			false

//**********************
// All runtime errors are fatal.
//

void Error (const char* msg) {
	NOTIFY_PROGRAMERROR(msg);
	CLEANUP_FUNCTION();
}

void Stop (unsigned line) {
	NOTIFY_PROGRAMSTOP(line);
	CLEANUP_FUNCTION();
}

void Error (const string& s) 
	{ Error(s.c_str()); }

///////////////////////////////////////////////////////

void input (const char* s, int* x) {
	if (s)
		OUTPUT_FUNCTION("%s (integer):", s);
	char* buff = SkipSpaces(ReadInteger(SkipSpaces(ReadInputString()), x));
	if (*buff)
		Error("Integer expected in input!");
}

void input (const char* s, int& x) { input(s, &x); }

//***********************

void input (const char* s, float* x) {
	if (s)
		OUTPUT_FUNCTION("%s (real):", s);
	if (!ReadReal(SkipSpaces(ReadInputString()), x))
		Error("Real expected in input!");
}

void input (const char* s, float& x) { input(s, &x); }

//***********************

void input (const char* s, string* x) {
	if (s)
		OUTPUT_FUNCTION("%s (string):", s);
	*x = ReadInputString();
}

void input (const char* s, char* x) {
	if (s)
		OUTPUT_FUNCTION("%s (string):", s);
	strcpy(x, ReadInputString());
}

//***********************

void input (const char* s, bool* x) {
	if (s)
		OUTPUT_FUNCTION("%s (boolean):", s);
	static char b[512];
	ReadInputString(b, sizeof(b)-1);
	StringToUpper(b);
	if (strcmp(b, TRUESTR) && strcmp(b, FALSESTR))
		Error("TRUE or FALSE expected for boolean input!");
	*x = !strcmp(b, TRUESTR);
}

void input (const char* s, char& x) {
	if (s)
		OUTPUT_FUNCTION("%s (character):", s);
	x = ReadInputString()[0];
}

//***********************

class Object;
void input (const char* s, Object** x);

#define INPUT(x)	input(#x, x)
#define	INPUTEX(x)	input((char*) 0, x)

///////////////////////////////////////////////////////

unsigned integerSpaces = 0;
#define	MAX_INTEGER_SPACES 10

void CTRL_OUTPUTEX_INTEGER (unsigned spaces) {
	if (spaces > MAX_INTEGER_SPACES)
		Error(MakeString("Spaces for INTEGER printing must be <= %d!\n", MAX_INTEGER_SPACES));
	integerSpaces = spaces;
}

void output (const char* s, int x) {
	OUTPUT_FUNCTION("%s: (integer) %d\n", s, x); 
}

void outputex (int x) {
	static char format[16];
	if (integerSpaces)
		sprintf(format,"%%%dd", integerSpaces);
	else
		strcpy(format, "%d");
	OUTPUT_FUNCTION(format, x); 
}

void output (const char* s, float x)
	{ OUTPUT_FUNCTION("%s:(real) %6.6f\n", s, x); }

void outputex (float x) { OUTPUT_FUNCTION("%6.6f", x); }

void output (const char* s, const string& x)
{ OUTPUT_FUNCTION("%s: (string) \"%s\"\n", s, x.c_str()); }

bool stringWithQuotes = true;
void CTRL_OUTPUTEX_STRING (bool val) {
	stringWithQuotes = val;
}

void outputex (const string& x) { 
	if (stringWithQuotes)
		OUTPUT_FUNCTION("\"%s\"", x.c_str()); 
	else
		OUTPUT_FUNCTION("%s", x.c_str()); 
}

void output (const char* s, bool x)
	{ OUTPUT_FUNCTION("%s: (boolean) \"%s\"\n", s, x ? "TRUE" : "FALSE"); }

void outputex (bool x) { OUTPUT_FUNCTION("\"%s\"", x ? "TRUE" : "FALSE"); }

void output (const char* s, char c)
	{ OUTPUT_FUNCTION("%s: (character) '%c'\n", s, c); }

//**************************
// The following covers the case where a pure 
// string message is displayed.
//
void output (const char* unused, const char* s) {
	OUTPUT_FUNCTION("%s\n", s);
}

void outputex (const char* s) {
	OUTPUT_FUNCTION("%s", s); 
}

void output (const char* s, Object* obj);

void outputex (Object* obj);

#define	LINE			"\n"
#define OUTPUT(x)		output(#x, x)
#define OUTPUTEX(x)		outputex(x)

//-------------------------------------------------------------------
// The automatic memory manager of the servant.
//
#define	PAGE_ROWS		60
#define	MAX_PAGES		300
#define	FIRST_TEMP_PAGE	150	// From this onwards it is for temporaries only!

inline bool IsValidAddress (unsigned p, unsigned r)
	{ return p < MAX_PAGES && r < PAGE_ROWS; }

class Object;
typedef pair<unsigned, Object*> Entry;

extern unsigned MemMakeIndex (unsigned page, unsigned row);

/////////////////////////////////////////////////////////////////////

class ObjectHolder {

	typedef map<unsigned, Object*> ObjectMap;
	ObjectMap objects;

	typedef map<unsigned, unsigned> Allocations;
	Allocations allocations;

	class DestroyObject : public unary_function<Object*, void> {
		public:
		void operator()(Object* p) const;
	};

	class RegisterForDestroy : public unary_function<Entry, void> {
		std::list<Object*>& objects;
		public:
		void operator()(const Entry& p) const;
		RegisterForDestroy (std::list<Object*>& l): objects(l){}
	};

	//**************************
	// Returns the beginning of a free continuous memory block
	// of the requested size.
	//
	const pair<unsigned, unsigned> GetBlock (unsigned size) {
		unsigned	sp, sr, n=0;
		bool		marked = false;
		for (unsigned i=0; i < MAX_PAGES; ++i)
			for (unsigned j=0; j<PAGE_ROWS; ++j)
				if (!IsOccupied(i, j)) {
					if (!marked)
						sp = i, sr = j, marked = true;
					if (++n == size)
						return pair<unsigned, unsigned>(sp, sr);
				}
				else
				if (marked)	// Failed to find a continuous region.
					marked = false, n = 0;

		return pair<unsigned, unsigned>(MAX_PAGES, PAGE_ROWS);
	}

	//**************************

	public:
	const pair<unsigned, unsigned> GetFirstAvailPos (unsigned fromPage, unsigned toPage) {
		for (unsigned i = fromPage; i <= toPage; ++i)
			for (unsigned j=0; j < PAGE_ROWS; ++j)
				if (!IsOccupied(i, j))
					return pair<unsigned, unsigned>(i, j);
		return pair<unsigned, unsigned>(MAX_PAGES, PAGE_ROWS);
	}

	// Gets a default memory position (non-temp).
	//
	const pair<unsigned, unsigned> GetFirstAvailPos (void) {
		return GetFirstAvailPos(0, FIRST_TEMP_PAGE-1);
	}

	const pair<unsigned, unsigned> GetAvailTempPos (void) {
		return GetFirstAvailPos(FIRST_TEMP_PAGE, MAX_PAGES-1);
	}

	//**************************
	// Allocate the memory requested. Register this allocation (we
	// need that to check disposal later).
	//
	const pair<unsigned, unsigned>	Grant (unsigned size) {

		pair<unsigned, unsigned> start = GetBlock(size);

		if (IsValidAddress(start.first, start.second)) {

			allocations[MemMakeIndex(start.first, start.second)] = size;

			unsigned p, r;
			for (p = start.first, r = start.second; size; ++r, --size) {
				if (r==PAGE_ROWS)
					r = 0, ++p;
				GetObjectAndCreateIfEmpty(p, r);
			}
		}

		return start;
	}

	//**************************
	// Free the dynamically allocated memory .
	// We firstly check that the starting address has been registered
	// earlier as a dynamic block.
	//
	void Release (const pair<unsigned, unsigned>& start);

	//**************************

	Object*	GetObjectAndCreateIfEmpty (unsigned page, unsigned row);

	Object&	GetExistingObject (unsigned page, unsigned row) {

		ObjectMap::iterator i = objects.find(MemMakeIndex(page, row));
		assert(i != objects.end());
		assert(i->second);
		return *i->second;
	}

	bool	IsOccupied (unsigned page, unsigned row) const
				{ return objects.find(MemMakeIndex(page, row)) != objects.end(); }

	// This is used for explicitly delcared variables or for implicitly 
	// instantiated temporary objects. It simply adds the object instances
	// or removes it (but no allocation is made).
	//
	void	ReservePosition (unsigned page, unsigned row, Object* obj = (Object*) 0) {
				assert(objects.find(MemMakeIndex(page, row))==objects.end());
				objects[MemMakeIndex(page, row)] = obj;
			}

	void	FreePosition (unsigned page, unsigned row) {
				assert(objects.find(MemMakeIndex(page, row))!=objects.end());
				objects.erase(MemMakeIndex(page, row));
				assert(objects.find(MemMakeIndex(page, row))==objects.end());
			}

	// We cannot iterate the original 'objects' list and delete, since upon
	// destructions objects remove them selves from the list.
	//
	~ObjectHolder() {

		// Round 1: register dor destruction in a list.
		//
		std::list<Object*> deleteList;
		std::for_each (objects.begin(), objects.end(), RegisterForDestroy(deleteList));

		// Round 2: massively delete.
		//
		std::for_each (deleteList.begin(), deleteList.end(), DestroyObject());

		objects.clear();
		deleteList.clear();
	}
};

static ObjectHolder memory;

//-------------------------------------------------------------------

#define	USEPOSITION(x)	\
	LineNameRecorder lineInfo##x(__LINE__, #x);	\
	Object& x = VerifyUnusedPosition

#define	USEREGION(x)	LineNameRecorder lineInfo##x(__LINE__, #x);	Region x
#define	TO				,

Object& POSITION (int page, int row);
Object&	POSITION (const Object& page, int row);
Object&	POSITION (int page, const Object& row);
Object&	POSITION (const Object& page, const Object& row);

//-------------------------------------------------------------------
// Raw memory usage support.
//
class Region {
	private:
	unsigned page, startRow, endRow;

	public:

	Object& operator[](const Object& i) const;

	Object& operator[](unsigned index) const {

		NOTIFY_REGIONINDEX(page, startRow, endRow, index);

		if (index > endRow-startRow)
			Error(
				MakeString(
					"[%d] for REGION is outside REGION(%d,%d...%d)!", 
					index,
					page,
					startRow,
					endRow
				)
			);
		return POSITION(page, startRow+index);
	}

	Region (unsigned p, unsigned sr, unsigned er);
};

//-------------------------------------------------------------------
// The basic memory type is an object class which can hold all types.
// 

#define TOTAL_TYPES	7

enum object_flip_t { 
	int_flip_t		= 0, 
	string_flip_t	= 1, 
	float_flip_t	= 2, 
	char_flip_t		= 3, 
	bool_flip_t		= 4, 
	address_flip_t	= 5, 
	undef_flip_t	= 6 
};

void VerifyType (const Object& obj, object_flip_t t);

#define TOTAL_STORAGES	3
enum storage_flip_t { 
		temp_flip_t	= 0, 
		var_flip_t	= 1, 
		mem_flip_t	= 2 
};

///////////////////////////////////////////////////////
// IN THIS CLASS I SHOULD NOT CALL EXPLICITLY THE CONSTRUCTOR
// AS THIS GENERATES DUPLICATE REGISTRATIONS IN THE NOTEBOOK MEMORY.
//
class Object {

	friend class Region;

	protected:
	object_flip_t	t;

	int				i;
	string			s;
	float			f;
	char			c;
	bool			b;
	Object*			a;

	static char* typeStrings[TOTAL_TYPES];
	static char* storageStrings[TOTAL_STORAGES];

	string			name;
	string			id;
	unsigned		page, row;
	storage_flip_t 	storage;

	//***********************

#define EMPTY_NAME "-"

#define	NOTIFY_DECLARATION()	\
	NOTIFY_VARDECL(				\
			name.c_str(),		\
			GetValue(this),		\
			GetTypeString(),	\
			GetStorageString(),	\
			page,				\
			row					\
	)

	void ConstructTemp (void) {
		pair<unsigned, unsigned> pos = memory.GetAvailTempPos();
		if (!IsValidAddress(pos.first, pos.second))
			Error("No more memory left on the NOTEBOOK for temporaries!");
		id = MakeString("POSITION(%d, %d):AUTO", page = pos.first, row = pos.second);
		memory.ReservePosition(page, row, this);
		storage = temp_flip_t;
		name = EMPTY_NAME;

		NOTIFY_DECLARATION();
	}

	void ConstructMem (unsigned p, unsigned d) {
		id = MakeString("POSITION(%d, %d):MEM", page = p, row = d);
		storage = mem_flip_t;
		name = EMPTY_NAME;

		NOTIFY_DECLARATION();
	}

	void ConstructVar (void) {
		pair<unsigned, unsigned> pos = memory.GetFirstAvailPos();
		if (!IsValidAddress(pos.first, pos.second))
			Error("No more memory left on the NOTEBOOK for variables!");
		id = MakeString("POSITION(%d, %d):VAR,%s", page=pos.first, row=pos.second, ts(t));
		memory.ReservePosition(page, row, this);
		storage = var_flip_t;
		name = EMPTY_NAME;

		NOTIFY_DECLARATION();
	}

	public:
	void SetName (const string& _name) {
		name = _name;
		NOTIFY_DECLARATION();
	}

	const char* GetName (void) const { return name.c_str(); }

	static const string GetValue (Object *x){
		static char value[1024];
		switch(x->t) {
			case int_flip_t :		sprintf(value, "%d", x->i);	break;
			case string_flip_t :	sprintf(value, "%s", x->s.c_str());	break;
			case float_flip_t :		sprintf(value, "%f", x->f);	break;
			case char_flip_t :		sprintf(value, "%c", x->c);	break;
			case bool_flip_t :		sprintf(value, "%s", x->b ? "TRUE":"FALSE");	break;
			case address_flip_t :	if(x->a)
										sprintf(value, "(%d, %d)", x->a->GetPage(), x->a->GetRow());
									else
										sprintf(value, "uninitialized address");
									break;
			case undef_flip_t :		sprintf(value, "UNDEFINED VALUE");	break;

			default:				assert(false);
		}

		return value;
	}

	//***********************

	public:

	unsigned SizeOf (void) const { return 1; }

	static const char*	ts(object_flip_t t) 
		{ return typeStrings[(unsigned) t]; }

	virtual const Object operator=(const Object& x) { 

		// If strongly typed variable and source type is different.
		//
		if (t != undef_flip_t && storage == var_flip_t && x.t != t) {	

			if (IsNumber() && x.IsNumber()) {	// If it is the case of numbers.
				if (t == int_flip_t)
					i = x.IntNumber();
				else
					f = x.FloatNumber();
				
				NOTIFY_VARWRITE(GetValue(this), GetTypeString(), page, row);
				return *this;
			}

			Error(
				MakeString(
					"Value assigned to '%s' is not of '%s' type!", 
					GetId().c_str(), 
					ts(t)
				)
			);	
			return *this;	//never gets here, just to avoid compiler warning
		}
		else {
			i = x.i, s = x.s, f = x.f, b = x.b, c = x.c, t = x.t, a = x.a;

			NOTIFY_VARWRITE(GetValue(this), GetTypeString(), page, row);
			return *this; 
		}
	}

	bool	IsNumber (void) const { return t == int_flip_t || t == float_flip_t; }
	int		IntNumber (void) const { assert(IsNumber()); return t==int_flip_t ? i : (int) f; }
	float	FloatNumber (void) const { assert(IsNumber()); return t==int_flip_t ? i : f; }

	const string	GetId (void) const { return id; }
	unsigned		GetRow (void) const { return row; }
	unsigned		GetPage (void) const { return page; }
	object_flip_t	GetType (void) const { return t; }
	Object*			GetAddressedObject (void) const { return a; }
	storage_flip_t	GetStorage (void) const { return storage; }

	const char*		GetTypeString (void) const { 
						assert(((unsigned) t) < TOTAL_TYPES);
							return typeStrings[(unsigned) t]; 
					}

	const char*		GetStorageString (void) const { 
						assert(((unsigned) storage) < TOTAL_STORAGES);
							return storageStrings[(unsigned) storage]; 
					}

	const string	GetStringValue (void) const {
						VerifyType(*this, string_flip_t);
						return s;
					}

	//***********************
	// Arithemtic operators, applicable only for numbers.
	//

#define	ARITH_OPERATOR(_op, _check)											\
	const Object operator _op(const Object& x) const {						\
	 NOTIFY_ARITHOP(#_op, page, row, x.page, x.row);						\
		if (!IsNumber() || !x.IsNumber())									\
			Error(MakeString("can use '%s' only for numbers!", #_op));		\
		_check																\
		if (t==float_flip_t || x.t==float_flip_t)							\
			return Object(FloatNumber() _op x.FloatNumber());				\
		else																\
			return Object(IntNumber() _op x.IntNumber());					\
	}

#define	CHECK_EMPTY
#define	CHECK_ZERO \
	if (x.FloatNumber() == 0.0) \
		Error("Division by zero!");
	
	ARITH_OPERATOR(+,CHECK_EMPTY)
	ARITH_OPERATOR(-,CHECK_EMPTY)
	ARITH_OPERATOR(*,CHECK_EMPTY)
	ARITH_OPERATOR(/,CHECK_ZERO)

	const Object operator%(const Object& x) const {

		NOTIFY_ARITHOP("%", page, row, x.page, x.row);

		if (t!=int_flip_t || x.t != int_flip_t)
			Error("Modulo '%' allowed only among integers!");
		if (x.i == 0)
			Error("Modulo with zero!");
		return Object(IntNumber() % x.IntNumber());
	}
	
#define UNARY_OPERATOR(_op)											\
	const Object operator _op(void) const {							\
		NOTIFY_UNARYOP(#_op, page, row);							\
		if (!IsNumber())											\
			Error("Unary '" #_op "' allowed only for numbers!");	\
		if (t == int_flip_t)										\
			return Object(_op IntNumber());							\
		else														\
			return Object(_op FloatNumber());						\
	}

	UNARY_OPERATOR(+)
	UNARY_OPERATOR(-)

	//***********************
	// Relational operators implemented for strings.
	//
#define RELAT_OPERATOR_STRING(_op)	\
	bool operator _op (const char* s2) const\
	{\
		VerifyType(*this, string_flip_t);\
		return strcmp(s.c_str(), s2) _op 0;\
	}

	RELAT_OPERATOR_STRING(>)
	RELAT_OPERATOR_STRING(>=)
	RELAT_OPERATOR_STRING(<)
	RELAT_OPERATOR_STRING(<=)
	RELAT_OPERATOR_STRING(==)
	RELAT_OPERATOR_STRING(!=)

	//***********************
	// Relational operators implemented for numbers.
	//
#define	RELAT_OPERATOR_NUMBER(_op, _ctype, _fliptype) \
	const bool operator _op(_ctype x) const { \
		if (IsNumber()) \
			return FloatNumber() _op x; \
		else { \
			Error(MakeString("Operation '%s %s %s' is not allowed!", ts(t), #_op, ts(_fliptype))); \
			return false; \
		} \
	}

	RELAT_OPERATOR_NUMBER(>, int, int_flip_t)
	RELAT_OPERATOR_NUMBER(>, double, int_flip_t)
	RELAT_OPERATOR_NUMBER(>=, int, int_flip_t)
	RELAT_OPERATOR_NUMBER(>=, double, int_flip_t)
	RELAT_OPERATOR_NUMBER(<=, int, int_flip_t)
	RELAT_OPERATOR_NUMBER(<=, double, int_flip_t)
	RELAT_OPERATOR_NUMBER(<, int, int_flip_t)
	RELAT_OPERATOR_NUMBER(<, double, int_flip_t)
	RELAT_OPERATOR_NUMBER(==, int, int_flip_t)
	RELAT_OPERATOR_NUMBER(==, double, int_flip_t)
	RELAT_OPERATOR_NUMBER(!=, int, int_flip_t)
	RELAT_OPERATOR_NUMBER(!=, double, int_flip_t)

	//***********************
	// Relational operator among boolean variables.
	//
	const bool operator==(bool val) const 
		{ VerifyType(*this, bool_flip_t); return b==val; }

	const bool operator!=(bool val) const 
		{ VerifyType(*this, bool_flip_t); return b!=val; }

	//***********************
	// Relational operators implemented for characters.
	//
#define RELAT_OPERATOR_CHAR(_op) \
	bool operator _op(char _c) const \
		{ VerifyType(*this, char_flip_t); return c _op _c; }

	RELAT_OPERATOR_CHAR(>)
	RELAT_OPERATOR_CHAR(>=)
	RELAT_OPERATOR_CHAR(<)
	RELAT_OPERATOR_CHAR(<=)
	RELAT_OPERATOR_CHAR(==)
	RELAT_OPERATOR_CHAR(!=)

	//***********************
	// Dispatcher of relational operators.
	//

#define	RELAT_OPERATOR_OBJECT(_op)								\
	const bool operator _op(const Object& x) const {		\
		NOTIFY_RELATOP(#_op, page, row, x.page, x.row);		\
		if (IsNumber() && x.IsNumber())						\
			return FloatNumber() _op x.FloatNumber();		\
		else												\
		if (t == string_flip_t && x.t == string_flip_t)		\
			return this->operator _op(x.s.c_str());			\
		else												\
		if (t == bool_flip_t && x.t == bool_flip_t)			\
			return this->operator _op(x.b);					\
		else												\
		if (t == char_flip_t && x.t == char_flip_t)			\
			return this->operator _op(x.c);					\
		else {												\
			Error(MakeString("Operation '%s %s %s' is not allowed!", ts(t), #_op, ts(x.t))); \
			return false;									\
		}													\
	}

	RELAT_OPERATOR_OBJECT(>)
	RELAT_OPERATOR_OBJECT(>=)
	RELAT_OPERATOR_OBJECT(<)
	RELAT_OPERATOR_OBJECT(<=)
	RELAT_OPERATOR_OBJECT(==)
	RELAT_OPERATOR_OBJECT(!=)

	//***********************

	void output_undefined (const char* msg) const {
		OUTPUT_FUNCTION("%s: (undefined) %s\n", msg, GetId().c_str());
	}

	void outputex_undefined (void) const {
		OUTPUT_FUNCTION("(undefined) %s", GetId().c_str());
	}

	void input_undefined (const char* msg) {

		OUTPUT_FUNCTION("%s, (undefined):", msg);
		char* buff = SkipSpaces(ReadInputString());

		if (!*SkipSpaces(ReadInteger(buff, &i))) {
			t = int_flip_t;
			if (ReadReal(buff, &f))
				t = float_flip_t;
		}
		else
		if (strlen(buff)==1)
			c = buff[0], t = char_flip_t;
		else {
			StringToUpper(buff);
			if (!strcmp(buff, TRUESTR))
				t = bool_flip_t, b = true;
			else
			if (!strcmp(buff, FALSESTR))
				t = bool_flip_t, b = false;
			else
				t = string_flip_t, s = buff;
		}
	}

	//***********************

	virtual void output (const char* msg) const {
		switch (t) {
			case int_flip_t	:	::output(msg, i); break;
			case float_flip_t :  ::output(msg, f); break;
			case string_flip_t : ::output(msg, s); break;
			case bool_flip_t :	::output(msg, b); break;
			case char_flip_t :	::output(msg, c); break;
			case address_flip_t:	::output(msg, a); break;
			case undef_flip_t:	output_undefined(msg); break;
			default: assert(false);
		}
	}

	virtual void outputex (void) const {
		switch (t) {
			case int_flip_t	:	::outputex(i); break;
			case float_flip_t :  ::outputex(f); break;
			case string_flip_t : ::outputex(s); break;
			case bool_flip_t :	::outputex(b); break;
			case char_flip_t :	::outputex(c); break;
			case address_flip_t:	::outputex(a); break;
			case undef_flip_t:	outputex_undefined(); break;
			default: assert(false);
		}
	}

	virtual void input (const char* msg) {
		switch (t) {
			case int_flip_t	:	::input(msg, &i); break;
			case float_flip_t : ::input(msg, &f); break;
			case string_flip_t : ::input(msg, &s); break;
			case bool_flip_t :	::input(msg, &b); break;
			case char_flip_t :	::input(msg, c); break;
			case address_flip_t:	::input(msg, &a); break;
			case undef_flip_t:	input_undefined(msg); break;
			default: assert(false);
		}

		NOTIFY_VARWRITE(GetValue(this), GetTypeString(), page, row);
	}

	//***********************
	// Implicit (automatically called by the compiler) constructors.
	//
	Object (int x) : i(x), t(int_flip_t),f(0), s(""), b(false), c('\0'), a((Object*) 0)
		{ ConstructTemp(); } 

	Object (double x) : f((float) x), t(float_flip_t), i(0), s(""), b(false), c('\0'), a((Object*) 0)
		{ ConstructTemp(); } 

	Object (bool x) : s(""), t(bool_flip_t),f(0), i(0), b(x), c('\0'), a((Object*) 0)
		{ ConstructTemp(); } 

	Object (char x) : s(""), t(char_flip_t),f(0), i(0), b(false), c(x), a((Object*) 0)
		{ ConstructTemp(); } 

	Object (const char* x) : s(x), t(string_flip_t),f(0), i(0), b(false), c('\0'), a((Object*) 0)
		{ ConstructTemp(); } 

	Object (const std::string& x) : s(x), t(string_flip_t), f(0), i(0), b(false), c('\0'), a((Object*) 0)
		{ ConstructTemp(); } 

	//***********************
	// The copy constructor.
	//
	Object (const Object& x) : i(x.i),s(x.s), f(x.f), b(x.b), c(x.c), t(x.t), a(x.a)
		{ ConstructTemp(); } 

	Object (void) : t(undef_flip_t), i(0), f(0), s(""), b(false), c('\0'), a((Object*) 0)
		{ ConstructTemp(); }

	//***********************
	// Explicit constructors.
	//
	Object (unsigned page, unsigned row) : t(undef_flip_t), i(0), f(0), s(""), b(false), c('\0'), a((Object*) 0)
		{ ConstructMem(page, row); }

	Object (object_flip_t _t) : t(_t), i(0), f(0), s(""), b(false), c('\0'), a((Object*) 0)
		{ ConstructVar(); }

	~Object() {
		NOTIFY_VARDESTR(page, row);
		memory.FreePosition(page, row);
	}
};

//*************************
// Asserts that (page, row) is a non-occupied memory position.
// If it does, it is a runtime error.
//
Object& VerifyUnusedPosition (unsigned page, unsigned row) {
	if (memory.IsOccupied(page, row))
		Error(
			MakeString(
				"POSITION(%d,%d) already in use, type '%s'!", 
				page, 
				row,
				memory.GetExistingObject(page, row).GetTypeString()
			)
		);
	return POSITION(page, row);
}

//*************************
// Region indexing with an object as argument.
//
Object& Region::operator[](const Object& i) const {

	if (i.t != int_flip_t)
		Error(MakeString("[%s] for REGION: integer expected!", i.GetTypeString()));

	int index = i.i;

	if (index > (int) (endRow-startRow))
		Error(
			MakeString(
				"[%d] for REGION is outside REGION(%d,%d TO %d)!",
				index,
				page,
				startRow,
				endRow
			)
		);

	return POSITION(page, startRow+index);
}

//***********************
// Because of the delete SetName usage this function must be defined
// after 'Object' class is defined.
//
Region::Region (unsigned p, unsigned sr, unsigned er) : page(p), startRow(sr), endRow(er) {

	NOTIFY_REGIONUSE(p, sr, er, LineNameRecorder::GetLine());

	if (sr > er)
		Error(MakeString("In REGION(%d,%d TO %d): start row > end row!",  p, sr, er));

	if (!IsValidAddress(p, sr) || !IsValidAddress(p, er))
		Error(MakeString("Wrong page / row values in REGION(%d,%d TO %d)!", p, sr, er));

	// Now create the objects that have not been created.
	//
	const char *name = LineNameRecorder::GetName();

	for (unsigned pos = 0; sr <= er; ++pos)
		VerifyUnusedPosition(page, sr++).SetName(
			MakeString("%s[%d]", name, pos)
		);
}

//***********************
// Because of the delete operator this function must be defined
// after 'Object' class is defined.
//
void ObjectHolder::Release (const pair<unsigned, unsigned>& start) {

	Allocations::iterator pos = allocations.find(MemMakeIndex(start.first, start.second));
	if (pos == allocations.end())
		Error(MakeString("No granted positions that start at (%d,%d)", start.first, start.second));

	unsigned size = pos->second;
	unsigned p, r;

	for (p = start.first, r = start.second; size; ++r, --size) {
		if (r==PAGE_ROWS)
			r = 0, ++p;
		Object* obj = GetObjectAndCreateIfEmpty(p, r);
		delete obj;
	}
}

//***********************

static Object& PositionWithNameExtraction (unsigned row, unsigned col) {

	Object *obj = memory.GetObjectAndCreateIfEmpty(row, col);

	if (LineNameRecorder::NotConsumed())
		obj->SetName(LineNameRecorder::GetName());

	return *obj;
}

Object& POSITION (const Object& i, const Object& j) {
	if (!i.IsNumber() || !j.IsNumber())
		Error(MakeString("Use of POSITION(%s,%s) is not allowed!", i.GetTypeString(), j.GetTypeString()));

	if (i.IntNumber() < 0 || j.IntNumber()<0)
		Error(MakeString("Invalid POSITION(%d,%d)!", i.IntNumber(), j.IntNumber()));

	return PositionWithNameExtraction(i.IntNumber(), j.IntNumber());
}

Object& POSITION (int i, int j) {
	if (i<0 || j<0)
		Error(MakeString("Invalid NOTEBOOK cell (%d,%d)!", i, j));

	return PositionWithNameExtraction((unsigned) i, (unsigned) j);
}

Object& POSITION (const Object& i, int j) {
	if (!i.IsNumber())
		Error(MakeString("Use of POSITION(%s,number) is not allowed!", i.GetTypeString()));
	if (i.IntNumber()<0 || j<0)
		Error(MakeString("Invalid POSITION(%d,%d)!", i.IntNumber(), j));
	return PositionWithNameExtraction(i.IntNumber(), (unsigned) j);
}

Object& POSITION (int i, const Object& j) {
	if (!j.IsNumber())
		Error(MakeString("Use of POSITION(number, %s) is not allowed!", j.GetTypeString()));

	if (i < 0 || j.IntNumber()<0)
		Error(MakeString("Invalid POSITION(%d,%d)!", i, j.IntNumber()));

	return PositionWithNameExtraction((unsigned) i, j.IntNumber());
}

//***********************
// This is how we cope with operator use when the native types
// are engaged first.
//
#define	NATIVE_OPERATOR(_type, _op, _res) \
const _res operator _op(_type x, const Object& y) { \
	Object temp(x); \
	return temp.operator _op(y); \
}

NATIVE_OPERATOR(int,+,Object)
NATIVE_OPERATOR(int,*,Object)
NATIVE_OPERATOR(int,-,Object)
NATIVE_OPERATOR(int,/,Object)
NATIVE_OPERATOR(int,%,Object)

NATIVE_OPERATOR(int,>,bool)
NATIVE_OPERATOR(int,>=,bool)
NATIVE_OPERATOR(int,<,bool)
NATIVE_OPERATOR(int,<=,bool)
NATIVE_OPERATOR(int,==,bool)
NATIVE_OPERATOR(int,!=,bool)

NATIVE_OPERATOR(double,+,Object)
NATIVE_OPERATOR(double,*,Object)
NATIVE_OPERATOR(double,-,Object)
NATIVE_OPERATOR(double,/,Object)
NATIVE_OPERATOR(double,%,Object)

NATIVE_OPERATOR(double,>,bool)
NATIVE_OPERATOR(double,>=,bool)
NATIVE_OPERATOR(double,<,bool)
NATIVE_OPERATOR(double,<=,bool)
NATIVE_OPERATOR(double,==,bool)
NATIVE_OPERATOR(double,!=,bool)

NATIVE_OPERATOR(const char*,>,bool)
NATIVE_OPERATOR(const char*,>=,bool)
NATIVE_OPERATOR(const char*,<,bool)
NATIVE_OPERATOR(const char*,<=,bool)
NATIVE_OPERATOR(const char*,==,bool)
NATIVE_OPERATOR(const char*,!=,bool)

NATIVE_OPERATOR(bool,==,bool)
NATIVE_OPERATOR(bool,!=,bool)

NATIVE_OPERATOR(char,==,bool)
NATIVE_OPERATOR(char,!=,bool)

//***********************

void output (const char* s, const Object& x) 
	{ x.output(s); }

void outputex (const Object& x) { x.outputex(); }

void input (const char* s, Object& x) 
	{ x.input(s); }

char* Object::typeStrings[TOTAL_TYPES] = {
	"INTEGER", "STRING", "REAL", "CHARACTER", "BOOLEAN", "ADDRESS", "undefined"
};

char* Object::storageStrings[TOTAL_STORAGES] = {
	"TEMP", "VAR", "MEM"
};

#define GETINPUT	getinput("GETINPUT")
#define GETINPUTEX	getinput((char*) 0)

const Object getinput (const char* msg) {
	Object obj;
	input(msg, obj);
	return obj;
}

///////////////////////////////////////////////////////
// Type specific derivatives of the Object base class
// for the variables,
//
void VerifyType (const Object& obj, object_flip_t t) {
	if (obj.GetType() != t)
		Error(MakeString("Type %s expected for %s!", Object::ts(t), obj.GetId().c_str()));
}

void VerifyNumber (const Object& obj) {
	if (obj.GetType() != int_flip_t && obj.GetType() != float_flip_t )
		Error(MakeString("Type %s, but not a number, for %s!", Object::ts(obj.GetType()), obj.GetId().c_str()));
}

/////////////////////////////////////////////////////////////////////
// Readable memory access support for external visualiser.
//
const char* MemGetObjectStorageString (unsigned page, unsigned row) {
	Object& obj = memory.GetExistingObject(page, row);
	return obj.GetStorageString();
}

const char* MemGetObjectTypeString (unsigned page, unsigned row) {
	Object& obj = memory.GetExistingObject(page, row);
	return obj.GetTypeString();
}

const char* MemGetObjectName (unsigned page, unsigned row) {
	Object& obj = memory.GetExistingObject(page, row);
	return obj.GetName();
}

unsigned MemMakeIndex (unsigned page, unsigned row)
		{ return PAGE_ROWS * page + row; }

//-------------------------------------------------------------------
// Detailed data types.
//
class IntObject : public Object {
	public:	
	const Object operator=(int x) 
		{ Object::operator=(x); return *this; }

	const Object operator=(const Object& x) 
		{ Object::operator=(x); return *this; }

	const Object operator=(double x) 
		{ Object::operator=((int) x); return *this; }

	int GetIntValue (void) const { return i; }

	explicit IntObject (void) : Object(int_flip_t){}
	IntObject (int x) : Object(int_flip_t) { Object::operator=(x); }
	IntObject (double x) : Object(int_flip_t) { Object::operator=((int) x); }
	IntObject (const Object& obj) : Object(int_flip_t) { VerifyNumber(obj); Object::operator=(obj); }
};

//************************

class CharObject : public Object {
	public:	
	const Object operator=(const Object& x) 
		{ return Object::operator=(x); }

	const Object operator=(char x)
		{ Object::operator=(x); return *this; }

	char Value (void) const { return c; }

	explicit CharObject (void) : Object(char_flip_t){}
	CharObject (char x) : Object(char_flip_t) { Object::operator=(x);}
	CharObject (const Object& obj) : Object(char_flip_t) { 
		VerifyType(obj, char_flip_t); 
		Object::operator=(obj); 
	}
};

//************************

class StringObject : public Object {

	void CheckIndex (const IntObject& i) const {
		if (i.IntNumber() < 0 || i.IntNumber() >= Length())
			Error(MakeString("In STRING, operator [%d] out of bounds!", i.IntNumber()));	
	}

	public:	
	const Object operator=(const Object& x) 
		{ return Object::operator=(x); }

	const Object operator=(const char* x) 
		{ Object::operator=(x); return *this; }

	int Length (void) const 
		{ return strlen(s.c_str()); }

	void Set (const CharObject& c, const IntObject& i) { 
		CheckIndex(i);	
		s[i.IntNumber()] = c.Value(); 
	}

	char Get (const IntObject& i) const { 
		CheckIndex(i);
		return s[i.IntNumber()]; 
	}

	const string GetValue (void) const { return s; }

	explicit StringObject (void) : Object(string_flip_t){}
	StringObject (const char* x) : Object(string_flip_t) { Object::operator=(x); }
	StringObject (const Object& obj) : Object(string_flip_t){
		VerifyType(obj, string_flip_t); 
		Object::operator=(obj);
	}
};

//************************

class RealObject : public Object {
	public:	
	const Object operator=(const Object& x) 
		{ return Object::operator=(x); }

	const Object operator=(int x) 
		{ Object::operator=((double) x); return *this; }

	const Object operator=(double x) 
		{  Object::operator=(x); return *this; }

	explicit RealObject (void) : Object(float_flip_t){}
	RealObject (double x) : Object(float_flip_t) { Object::operator=( x);}
	RealObject (int x) : Object(float_flip_t) { Object::operator=((double) x);}
	RealObject (const Object& obj);
};

RealObject::RealObject (const Object& obj) { 
	VerifyNumber(obj); 
	t = float_flip_t;
	Object::operator=(obj); 
}

//************************

class BoolObject : public Object {
	public:	
	const BoolObject operator||(const BoolObject& x) const 
		{ return Object(b || x.b); }

	const BoolObject operator&&(const BoolObject& x) const
		{ return Object(b && x.b); }

	bool operator!(void) const
		{ return !b; }

	const Object operator=(const Object& x) 
		{ return Object::operator=(x); }
	const Object operator=(bool x) 
		{ Object::operator=(x); return *this; }

	explicit BoolObject (void) : Object(bool_flip_t){}
	BoolObject (bool x) : Object(bool_flip_t) { Object::operator=(x); }
	BoolObject (const Object& obj) : Object(bool_flip_t) { 
		VerifyType(obj, bool_flip_t); 
		Object::operator=(obj); 
	}
};

class CondValueTrapper {
private:
	std::string stmt;
	bool val;
public:
	operator bool() { 
		std::string s;
		if (val)
			s = "TRUE condition, entering ";
		else
			s = "FALSE condition, skipping ";
		s.append(stmt);
		NOTIFY_CONDITIONVALUE(s.c_str(), val);
		return val;
	}
	CondValueTrapper& operator,(bool b) 
			{ val = b; return *this; }
	CondValueTrapper& operator,(BOOLEAN& val) 
			{ this->operator,(val == true); return *this; }
	CondValueTrapper(const char* _stmt) : stmt(_stmt){}
};

//-------------------------------------------------------------------
// FLIP ARRAYS.
//

template <const unsigned N> class ArraySuper {
	private:
	pair<unsigned, unsigned> pos;

	public:
	ArraySuper (void) {
		pos = memory.GetFirstAvailPos();
		NOTIFY_ARRAYALLOC(N);	
	}
	~ArraySuper (){}

	unsigned GetPage (void) const { return pos.first; }
	unsigned GetRow (void) const { return pos.second; }
};

template <class T, const unsigned N> class Array : public ArraySuper<N> {

	private:
	T array[N];
	
	// We need to ensure that declaration notification takes place AFTER
	// construction of the array, and destruction notification BEFORE
	// its elements get destroyed.

	public:
	Array(void) {
		NOTIFY_ARRAYDECL(GetPage(), GetRow(), N);
	}

	~Array() {
		NOTIFY_ARRAYDESTR(GetPage(), GetRow(), N);
	}

	T& operator[](int i) {
		NOTIFY_ARRAYINDEX(GetPage(), GetRow(), N, i);

		if (i < 0 || i >= N)
			Error(MakeString("[%d] for ARRAY is outside array [0,%d) boundaries!", i, N));
		return array[i];
	}

	unsigned SizeOf (void) const { return N * array[0].SizeOf(); };

	private:

	// This is a way to filter-out implicit compiler conversions.
	// Putting explicitly the type and passing it as template argument
	// ensures perfect type matching. Underirable types are detected 
	// through functions that do not return a value as they should.
	//
	class TypeFilteringClass {

		public:

		/* Undesirable types. */
		int operator()(const STRING& s) const {}
		int operator()(const char* s) const {}
		int operator()(char c) const {}
		int operator()(const REAL& r) const {}
		int operator()(const CHARACTER& c) const {}
		int operator()(const BOOLEAN& b) const {}

		/* Desirable types. */
		int operator()(const INTEGER& i) const 
			{ return i.GetIntValue(); }

		int operator()(const Object& obj) const { 
			if (obj.GetType() != int_flip_t)
				Error("Index for ARRAY is not an integer!");
			return obj.IntNumber();
		}
	};

	public:
	template <class C_IntObject_Object>
	T& operator[](const C_IntObject_Object& obj) {
		return this->operator[](TypeFilteringClass()(obj));
	}

	void output (const char* msg) {
		for (unsigned i=0; i<N; ++i)
			array[i].output(MakeString("%s[%d]", msg, i).c_str());
	}

	void input (const char* msg) {
		for (unsigned i=0; i<N; ++i)
			array[i].input(MakeString("%s[%d]", msg, i).c_str());
	}
};

#define	ARRAY(t, n) Array< t, n>
#define	_array	ARRAY

/////////////////////////////////////////////////////////////////////

Object* ObjectHolder::GetObjectAndCreateIfEmpty (unsigned page, unsigned row) {

	if (!IsValidAddress(page, row))
		Error(MakeString("Out of bounds NOTEBOOK address (%d,%d)!", page, row));

	ObjectMap::iterator i = objects.find(MemMakeIndex(page, row));

	if (i==objects.end()) {
		Object* obj = new Object(page, row);
		objects[MemMakeIndex(page, row)] = obj;
		return obj;
	}
	else
		return i->second;
}

//***********************

void ObjectHolder::DestroyObject::operator()(Object* p) const { 
	delete p; 
}

void ObjectHolder::RegisterForDestroy::operator()(const Entry& p) const {
	if (p.second && p.second->GetStorage() == mem_flip_t)
		objects.push_back(p.second); 
}

//-------------------------------------------------------------------
// Addresses for pointers. Supports arithmetic with auto
// flipping to the next notebook page.
//
class AddressObject : public Object {
	
	private:
	const AddressObject Offset (int offset) const {
		if (!a)
			Error("Adding to an unitialised address var!");
		unsigned cp = a->GetPage() + (a->GetRow() + offset) / PAGE_ROWS;
		unsigned cr = (a->GetRow() + offset) % PAGE_ROWS;
		
		if (cp >= MAX_PAGES || cr >= PAGE_ROWS)
			Error(
				MakeString(
					"Added value %d on address(%d,%d) is out of NOTEBOOK!", 
					offset, 
					a->GetPage(), 
					a->GetRow()
				)
			);

		// Create a new address object and return it.
		//
		return AddressObject(memory.GetObjectAndCreateIfEmpty(cp, cr));
	}

	void SetPointedObject (Object* obj) { a = obj; }

	public:
	enum nil_flip_t { nil_flip_value = 0 };

	//********************
	// Address arithmetic.
	//
	const AddressObject operator+(const Object& obj) const {
		if (obj.GetType() != int_flip_t)
			Error("Adding to an address var something other than integer!");
		return Offset(obj.IntNumber());
	}

	const AddressObject operator-(const Object& obj) const {
		if (obj.GetType() != int_flip_t)
			Error("Subtracting from an address var something other than integer!");
		return Offset(-obj.IntNumber());
	}
	
	//********************
	// Equality operators.
	//
	bool operator==(nil_flip_t v) const
		{ return !GetAddressedObject(); }

	bool operator!=(nil_flip_t v) const
		{ return !!GetAddressedObject(); }

	bool operator==(const AddressObject& obj) const
		{ return GetAddressedObject()==obj.GetAddressedObject(); }

	bool operator!=(const AddressObject& obj) const
		{ return GetAddressedObject()!=obj.GetAddressedObject(); }

	const AddressObject operator=(const AddressObject& x)
		{ Object::operator=(x); return *this; }

	const AddressObject operator=(nil_flip_t v)
		{ SetPointedObject((Object*) 0); return *this; }

	// De-reference through array subscript.
	//
	Object& operator[](const Object& index) const;

	//********************
	// Constructors, allowing direct use of a specific memory address.
	//
	AddressObject (void);

	AddressObject (unsigned p, unsigned r) : Object(address_flip_t) { // Variable.
		if (!IsValidAddress(p, r))
			Error("Invalid NOTEBOOK memory address given to address var!");
		SetPointedObject(memory.GetObjectAndCreateIfEmpty(p, r));
	}

	AddressObject (Object* obj);

	AddressObject (const AddressObject& obj);

	AddressObject (nil_flip_t v)
		{ t = address_flip_t; SetPointedObject((Object*) 0); }
};

#define	NIL		AddressObject::nil_flip_value

//********************
// This is only used for temporary objects.
// Thats why we use the default construct for temporary objects.
//
AddressObject::AddressObject (Object* obj) {
	t = address_flip_t;
	SetPointedObject(obj);
}

AddressObject::AddressObject (void) : Object(address_flip_t) {}

AddressObject::AddressObject (const AddressObject& obj) : Object(address_flip_t) {
	a = obj.GetAddressedObject(); 
} 

//********************
// Typical array-style addressing for pointer types.
//

// Cannot use "#define CONTENTOF(x, y) _CONTENTOF(x, y, __LINE__)"
// since a macro cannot be overloaded to accept 1 or 2 parameters.
//
#define CONTENTOF	ContentOf(__LINE__)

Object& _CONTENTOF (const Object& x, unsigned line);

Object& AddressObject::operator[](const Object& index) const {
	VerifyType(index, int_flip_t);
	return _CONTENTOF(Offset(index.IntNumber()), 0);
}

//********************
// Address-of and memory de-reference operator emulation.
//

// We always narrow down to the basic address-using function.
//
Object& _CONTENTOF (unsigned page, unsigned row, unsigned line) {
	return _CONTENTOF(AddressObject(page, row), line);
}

Object& _CONTENTOF (const Object& x, unsigned line) {
	NOTIFY_CONTENTOF(x.GetPage(), x.GetRow(), line);
	if (x.GetType() != address_flip_t)
		Error(MakeString("CONTENTOF used for %s which is not an ADDRESS var!", x.GetId().c_str()));
	Object* obj = x.GetAddressedObject();
	if (!obj)
		Error(MakeString("Getting value of ADDRESS var %s that has wrong value!", x.GetId().c_str()));
	return *obj;
}

const AddressObject _POSITIONOF (const Object& x, unsigned line) {
	NOTIFY_POSITIONOF(x.GetPage(), x.GetRow(), line);
	return AddressObject(x.GetPage(), x.GetRow());
}

const AddressObject _POSITIONOF (unsigned page, unsigned row, unsigned line) {
	NOTIFY_POSITIONOF(page, row, line);
	return AddressObject(page, row);
}

template <class T>
int _PAGEOF (const T& x, unsigned line) {
	NOTIFY_PAGEOF(x.GetPage(), x.GetRow(), line);
	return x.GetPage();
}

template <class T>
int _ROWOF (const T& x, unsigned line) {
	NOTIFY_ROWOF(x.GetPage(), x.GetRow(), line);
	return x.GetRow();
}

//******************
// This template function may be used to bypass some compiler
// error in references.
//
template <class T> const T& _ref (T& r) { return r; }

#define POSITIONOF		PositionOfLineTrapper(__LINE__)
#define PAGEOF(x)		_PAGEOF(x, __LINE__)
#define ROWOF(x)		_ROWOF(x, __LINE__)

class PositionOfLineTrapper : public LineTrapper {
	public:
	const AddressObject operator()(const Object& x) const
			{ return(_POSITIONOF(x, line)); }

	const AddressObject operator()(unsigned page, unsigned row) const
			{ return(_POSITIONOF(page, row, line)); }

	PositionOfLineTrapper (unsigned _line) : LineTrapper(_line){}
};

//********************
// Dynamic memory allocation functions.
//

#define GRANTPOSITIONS(x, y)	_GRANTPOSITIONS(x, y, __LINE__)
#define RELEASEPOSITIONS(x)		_RELEASEPOSITIONS(x, __LINE__)

void _GRANTPOSITIONS (AddressObject& obj, int size, unsigned line) {
	if (size <= 0)
		Error(MakeString("Invalid allocation size %d", size));
	pair<unsigned, unsigned> start = memory.Grant(size);
	if (!IsValidAddress(start.first, start.second))
		Error(MakeString("Not enough NOTEBOOK space for %d lines", size));
	NOTIFY_GRANT(start.first, start.second, size, line);
	obj = AddressObject(start.first, start.second);
}

void _GRANTPOSITIONS (AddressObject& obj, const IntObject& x, unsigned line) {
	if (x.IntNumber() <= 0)
		Error("GRANTPOSITIONS needs a positive integer number!");
	_GRANTPOSITIONS(obj, x.IntNumber(), line);
}

void _RELEASEPOSITIONS (const AddressObject& addr, unsigned line) {
	unsigned page = addr.GetAddressedObject()->GetPage();
	unsigned row = addr.GetAddressedObject()->GetRow();
	NOTIFY_RELEASE(page, row, line);
	memory.Release(pair<unsigned, unsigned>(page, row));
}

//********************
// Address type objects can be defined through input 
// as well!
//
void input (const char* s, Object** x) {
	OUTPUT_FUNCTION("%s (page, row):", s);
	unsigned page, row;
	if (sscanf(ReadInputString(), "%d %d", &page, &row) != 2)
		Error("Address input error!");
	if (!IsValidAddress(page, row))
		Error(MakeString("Invalid address(%d,%d) was given!",page, row));
	*x = memory.GetObjectAndCreateIfEmpty(page, row);
}

void output (const char* s, Object* obj) {
	if (obj)
		OUTPUT_FUNCTION("%s: (address) page %d, row %d\n", s, obj->GetPage(), obj->GetRow());
	else
		OUTPUT_FUNCTION("%s: nil address!\n", s);
}

void outputex (Object* obj) {
	if (obj)
		OUTPUT_FUNCTION("(address) page %d, row %d", obj->GetPage(), obj->GetRow());
	else
		OUTPUT_FUNCTION("address var not initialised!");
}

//********************
// The special input / output functions for array types,
// taking care of the loops automatically.
//
template <class T, const unsigned N>
void output (const char* msg, Array<T,N>& p) { p.output(msg); }

template <class T, const unsigned N>
void input (const char* msg, Array<T,N>& p) { p.input(msg); }

//-------------------------------------------------------------------
// Address type for strongly typed pointers. Those actually resemble
// smart-pointers when you strip-off the garbage collection stuff from the
// later.
//
template <class T> class AddressObjectEx : public Object {
	
	private:
	T*			value;
	unsigned	size;
	int			pos;
	bool		isDynamic;

	//********************

	const AddressObjectEx Offset (int offset) const {

		if (!value)
			Error("Adding to an unitialised address var!");

		// Create a new address object and return it.
		//
		return AddressObjectEx(value, size, isDynamic, pos+offset);
	}

	AddressObjectEx (T* _value, unsigned _size, bool _isDynamic, int _pos = 0) :
		value(_value), size(_size), pos(_pos), isDynamic(_isDynamic){}

	public:

	//********************
	// Address arithmetic.
	//
	const AddressObjectEx operator+(const Object& obj) const {
		if (obj.GetType() != int_flip_t)
			Error("Adding to an address something not an integer!");
		return Offset(obj.IntNumber());
	}

	const AddressObjectEx operator-(const Object& obj) const {
		if (obj.GetType() != int_flip_t)
			Error("Subtracting from an address something not an integer!");
		return Offset(-obj.IntNumber());
	}
	
	//********************
	// Equality operators.
	//
	bool operator==(AddressObject::nil_flip_t v) const	
		{ return !value; }

	bool operator!=(AddressObject::nil_flip_t v) const
		{ return !this->operator==(v); }

	bool operator==(const AddressObjectEx& obj) const
		{ return value==obj.value && pos == obj.pos; }

	bool operator!=(const AddressObjectEx& obj) const
		{ return !this->operator==(obj); }

	const AddressObjectEx operator=(const AddressObjectEx& x)
		{ value = x.value; size = x.size; isDynamic = x.isDynamic; pos = x.pos; return *this; }

	const AddressObjectEx operator=(AddressObject::nil_flip_t v)
		{ value = (T*) 0; pos = 0; size = 0; isDynamic = false; return *this; }

	//********************
	// De-reference and array subscript.
	//
	T& operator*(void) const { 
		if (!value)
			Error("Getting content of 'NIL' address!");
		if (pos < 0 || pos >= size)
			Error(MakeString("Address offset %d is out of legal bounds [0,%d)!", pos, size));
		return value[pos];
	}

	T& operator[](const Object& index) const {	
		VerifyType(index, int_flip_t);
		return *Offset(index.IntNumber());
	}

	//********************
	// Allocation.
	//
	void Allocate (int _size) {
		if (_size <= 0)
			Error(MakeString("Invalid allocation size for ADDRESSEX variable %d!", _size));
		value = new T[size  = _size];
		isDynamic = true;
		pos = 0;
	}

	void Release (void) {
		if (!value)
			Error("Releasing a NIL address!");
		else
		if (!isDynamic)
			Error("Releasing memory not granted dynamically!");
		else
		if (pos)
			Error("Releasing using an address with offset!");
		delete[] value;
		value = (T*) 0;
		pos = 0;
		isDynamic = false;
	}

	//********************
	// Address-of and array to pointer conversion.
	//
	static const AddressObjectEx AddressOf (T& obj) 
		{ return AddressObjectEx(&obj, 1, false); }

	template <class C, const unsigned N>
	const AddressObjectEx operator=(Array<C,N>& x)
		{ value = &x[0]; pos = 0; size = N; isDynamic = false; return *this; }

	//********************
	// Overloaded output functions.
	//
	void output (const char* msg) const {
		if (!value)
			OUTPUT_FUNCTION("%s: (ADDRESSEX), NIL\n", msg);
		else {
			AddressObject addr;
			addr = _ref(_POSITIONOF(*((Object*) value), 0)); // Tricky casting, but works.
			OUTPUT_FUNCTION(
				"%s: (ADDRESSEX), (%d,%d), SIZE=%d, OFFSET=%d\n", 
				msg,
				addr.GetAddressedObject()->GetPage(), 
				addr.GetAddressedObject()->GetRow(),
				size, 
				pos
			);
		}
	}

	void outputex (void) const {
		if (!value)
			OUTPUT_FUNCTION("NIL");
		else
			_ref(_POSITIONOF(*((Object*) value), 0)).outputex();
	}

	void input (const char* msg) {
		Error("Cannot input a value for an ADDRESSEX variable!");
	}

	//********************
	// Constructors. Notice that for some inexplicable reason
	// the default constructor of the base class is not called, so
	// we do put explicitly Object(address_flip_t) to ensure a variable
	// is properly created.
	//
	AddressObjectEx (void) : 
		Object(address_flip_t), value((T*) 0), size(0), pos(0), isDynamic(false) {}

	AddressObjectEx (const AddressObjectEx& obj) : Object(address_flip_t)
		{ value = obj.value; pos = obj.pos; size = obj.size; isDynamic = obj.isDynamic; }

	AddressObjectEx (AddressObject::nil_flip_t v) : 
		Object(address_flip_t), value((T*) 0), size(0), pos(0), isDynamic(false) {}

	template <class C, const unsigned N>
	AddressObjectEx (Array<C,N>& x) : Object(address_flip_t) 
		{ value = &x[0]; pos = 0; size = N; isDynamic = false; }
};

///////////////////////////////////////////////////////

#define	ADDRESSEX(_type)	AddressObjectEx<_type>

template <class T>
T& CONTENTOFEX (AddressObjectEx<T>& x) { return *x; }

template <class T>
void _GRANTPOSITIONS (AddressObjectEx<T>& x, int size, unsigned line) { 
	x.Allocate(size);
	NOTIFY_GRANT((*x).GetPage(), (*x).GetRow(), size, line);
}

template <class T>
void _GRANTPOSITIONS (AddressObjectEx<T>& x, const IntObject& i, unsigned line)
	{ _GRANTPOSITIONS(x, i.IntNumber(), line); }

template <class T>
void _RELEASEPOSITIONS (AddressObjectEx<T>& x, unsigned line) { 
	NOTIFY_RELEASE((*x).GetPage(), (*x).GetRow(), line);
	x.Release();
}

template <class T>
const AddressObjectEx<T> POSITIONOFEX (T& x) {
	return AddressObjectEx<T>::AddressOf(x);
}

//-----------------------------------------------------
// Program and block markers. Those incorporate the runtime
// event triggers for the visualization linkage.
// 

INSTANTIATION_TRAPPER(_Program, NOTIFY_PROGRAMSTART())

#define PROGRAM_BEGIN						\
	int main (int argc, char** argv) {		\
		_servant_argc = argc;				\
		_servant_argv = argv;
		
#define PROGRAM_END				\
		NOTIFY_PROGRAMEND();	\
		return 0;				\
	}

#define BEGIN			{ NOTIFY_BLOCKENTER(__LINE__);
#define END				NOTIFY_BLOCKEXIT(__LINE__); }

#ifndef	__FUNCTION__
#define	__FUNCTION__	"<Unknown>"
#endif

#define FUNCBEGIN		{ NOTIFY_FUNCENTER(__FUNCTION__, __LINE__);
#define FUNCEND			NOTIFY_FUNCEXIT(__FUNCTION__, __LINE__); }
#define RESULT(x)		return (NOTIFY_RESULT(#x, __FUNCTION__, __LINE__), x)

#define	FINISH			do	\
						{	\
							NOTIFY_FINISH(__FUNCTION__, __LINE__);	\
							return;	\
						} while(false)

// The FINISH could employ also patterns similar for BREAK or CONTINUE.
//

#define	BREAK			if (true) {	NOTIFY_BREAK(__LINE__);	break; } else
#define	CONTINUE		if (true) {	NOTIFY_CONTINUE(__LINE__);	continue; } else
#define IF				if (IfTrapper(__LINE__), CondValueTrapper("IF"),

// Causes a syntax error if empty initial expression list 
// is given in FLIP.
//
#define	FOR				for (ForTrapper(__LINE__), 
#define	WHILE			while (WhileTrapper(__LINE__), CondValueTrapper("WHILE"),
#define	STOP			Stop(__LINE__)
#define THEN			)
#define	ELSE			else
#define	ACTION			))
#define	DO				)

///////////////////////////////////////////////////////////////
// Case statements.
//

class CaseEvalList {
	private:
	std::list<Object> values;

	public:
	CaseEvalList& operator,(const Object& obj)
		{ values.push_back(obj); return *this; }

	bool operator==(Object& obj) const 
		{ return std::find(values.begin(), values.end(), obj) != values.end(); }

	CaseEvalList (void){}
	~CaseEvalList(){ values.clear(); }
};

//********************

struct CaseFetch {
	Object& val;
	bool operator==(const CaseEvalList& l) 
		{ return l == val; }
	CaseFetch (Object& v) : val(v){}
};

//********************

#define CASEBEGIN(expr) \
	do { NOTIFY_CASEBEGIN(__LINE__); Object _caseVal(expr); if (false) ;

#define ON		else if (CaseFetch(_caseVal) == (CaseEvalList(),
#define CASEND	NOTIFY_CASEEND(__LINE__); } while(false)

#define	_on			ON
#define	_casend		CASEND

///////////////////////////////////////////////////////////////
// Records. Based on a super-class to trap starting address
// and to post notifications.
//
struct RecordSuper {
	private:
	pair<unsigned, unsigned> pos;

	public:
	RecordSuper (void) {
		pos = memory.GetFirstAvailPos();
		NOTIFY_RECORDDECL(); 
	}
	~RecordSuper ()	{ NOTIFY_RECORDDESTR(); }

	unsigned GetPage (void) const { return pos.first; }
	unsigned GetRow (void) const { return pos.second; }
};

struct FreeCellTrapper {
	pair<unsigned, unsigned> pos;
	FreeCellTrapper (void) { pos = memory.GetFirstAvailPos(); }
};

#define	RECORD		struct
#define	RECBEGIN	: public RecordSuper {

#define	RECEND	\
		FreeCellTrapper lastCell;	\
		unsigned SizeOf (void) const {	\
			unsigned end   = MemMakeIndex(lastCell.pos.first, lastCell.pos.second);	\
			unsigned start = MemMakeIndex(GetPage(), GetRow());	\
			return end - start;	\
		}	\
	}

///////////////////////////////////////////////////////////////

#define	SIZEOF(var)		((int) ((var).SizeOf()))
#define FUNCTION(t)		t
#define PROCEDURE		FUNCTION(void)
#define ARGUMENTS

#define	AND				&&
#define	OR				||
#define	NOT				!
#define	EQUAL			==
#define	DIFFERENT		!=

#define	CONST_INTEGER	const int
#define	CONST_REAL		const double
#define	CONST_STRING	const string
#define	CONST_CHARACTER	const char
#define	CONST_BOOLEAN	const bool

///////////////////////////////////////////////////////
// String manipulation functions.
//

#define string_length(x)	_string_length(x, __LINE__)
const INTEGER _string_length (const StringObject& str, unsigned line) { 
	NOTIFY_FUNCUSE("string", "string_length", line);
	return str.Length(); 
}

#define type_of(x)	_type_of(x, __LINE__)
const STRING _type_of (const Object& o, unsigned line) { 
	NOTIFY_FUNCUSE("types", "type_of", line);
	return o.GetTypeString(); 
}

#define string_get(x, y)	_string_get(x, y, __LINE__)
const CHARACTER _string_get (const StringObject& str, const IntObject& i, unsigned line)  { 
	NOTIFY_FUNCUSE("string", "string_get", line);
	return str.Get(i); 
}

#define string_isinteger(x)	_string_isinteger(x, __LINE__)
const BOOLEAN _string_isinteger (const StringObject& str, unsigned line)  { 
	NOTIFY_FUNCUSE("string", "string_isinteger", line);
	int i;
	if (!strlen(str.GetValue().c_str()))
		return false;
	else
		return !*SkipSpaces(ReadInteger(SkipSpaces(const_cast<char*>(str.GetValue().c_str())), &i));
}

#define string_isreal(x)	_string_isreal(x, __LINE__)
const BOOLEAN _string_isreal (const StringObject& str, unsigned line)  { 
	NOTIFY_FUNCUSE("string", "string_isreal", line);
	float f;
	if (!strlen(str.GetValue().c_str()))
		return false;
	else
		return !!ReadReal(SkipSpaces(const_cast<char*>(str.GetValue().c_str())), &f);
}

#define string_set(x, y, z)	_string_set(x, y, z, __LINE__)
void _string_set (StringObject& str, const IntObject& i, const CharObject& c, unsigned line) { 
	NOTIFY_FUNCUSE("string", "string_set", line);
	str.Set(c, i); 
}

///////////////////////////////////////////////////////
// Math functions. The implementation is not re-use
// optimised.
//

#define sqr(x)	_sqr(x, __LINE__)
const Object _sqr (const Object& x, unsigned line) {
	NOTIFY_FUNCUSE("math", "sqr", line);
	if (x.GetType() != int_flip_t && x.GetType() != float_flip_t)
		Error(MakeString("Invalid type '%s' passed to 'sqr'!", Object::ts(x.GetType())));
	if (x.GetType() == int_flip_t)
		return Object(x.IntNumber() * x.IntNumber());
	else
		return Object(x.FloatNumber() * x.FloatNumber());
}

//*********************

const Object _sqrt (const Object& x, unsigned line) {
	NOTIFY_FUNCUSE("math", "sqrt", line);
	if (x.GetType() != int_flip_t && x.GetType() != float_flip_t)
		Error(MakeString("Invalid type '%s' passed to 'sqrt'!", Object::ts(x.GetType())));
	return Object((float) sqrt(x.FloatNumber()));
}

// Should be defined here not to hide the native sqrt used above.
//
#define sqrt(x)	_sqrt(x, __LINE__)

//*********************
// Trigonometric functions.
// The value of PI is not as accurate as it can be.
//
#define PI				3.141628
#define TORADS(angle)	((2.0 * PI * (angle)) / 360.0)

#define	TRIGONOMETRIC_FUNC(f)										\
const Object _##f (const Object& x, unsigned line) {				\
	NOTIFY_FUNCUSE("trigonometric", #f, line);						\
	if (x.GetType() != int_flip_t && x.GetType() != float_flip_t)	\
		Error(														\
			MakeString(												\
				"Invalid type '%s' passed to '%s'!",				\
				x.GetTypeString(),									\
				#f													\
			)														\
		);															\
	return Object(f(TORADS(x.FloatNumber())));						\
}

TRIGONOMETRIC_FUNC(cos)
TRIGONOMETRIC_FUNC(sin)
TRIGONOMETRIC_FUNC(tan)

// Defined afterwards because the native functions are used above.
//
#define cos(x)	_cos(x, __LINE__)
#define sin(x)	_sin(x, __LINE__)
#define tan(x)	_tan(x, __LINE__)

///////////////////////////////////////////////////////
// Date functions.
//

#define curr_month()	_curr_month(__LINE__)
const INTEGER _curr_month (unsigned line) {
	NOTIFY_FUNCUSE("date", "curr_month", line);
	time_t t;
	time(&t);
	return gmtime(&t)->tm_mon;
}

//*********************

#define curr_day()	_curr_day(__LINE__)
const INTEGER _curr_day (unsigned line) {
	NOTIFY_FUNCUSE("date", "curr_day", line);
	time_t t;
	time(&t);
	return gmtime(&t)->tm_wday;
}

//*********************

#define curr_date()	_curr_date(__LINE__)
const INTEGER _curr_date (unsigned line) {
	NOTIFY_FUNCUSE("date", "curr_date", line);
	time_t t;
	time(&t);
	return gmtime(&t)->tm_mday;
}

//*********************

#define curr_year()	_curr_year(__LINE__)
const INTEGER _curr_year (unsigned line) {
	NOTIFY_FUNCUSE("date", "curr_year", line);
	time_t t;
	time(&t);
	return gmtime(&t)->tm_year + 1900;
}

///////////////////////////////////////////////////////

unsigned static ProgramStartLine (void) { return __LINE__; }

#define	_int				INTEGER
#define	_float				REAL
#define	_string				STRING
#define	_bool				BOOLEAN
#define	_char				CHARACTER
#define	_address			ADDRESS
#define	_record				RECORD
#define	_recbegin			RECBEGIN
#define	_recend				RECEND

#define	_input				INPUT
#define	_inputex			INPUTEX
#define	_line				LINE
#define	_output				OUTPUT
#define	_outputex			OUTPUTEX
#define _getinput			GETINPUT
#define _getinputex			GETINPUTEX

#define	_position			POSITION
#define	_useposition		USEPOSITION
#define	_useregion			USEREGION
#define	_to					TO
#define	_nil				NIL
#define	_positionof			POSITIONOF
#define	_pageof				PAGEOF
#define	_rowof				ROWOF
#define	_contentof			CONTENTOF
#define	_grantpositions		GRANTPOSITIONS
#define	_releasepositions	RELEASEPOSITIONS
#define	_addressex			ADDRESSEX
#define	_contentofex		CONTENTOFEX

#define	_progam_begin		PROGRAM_BEGIN
#define	_program_end		PROGRAM_END
#define	_begin				BEGIN
#define	_end				END

#define	_function			FUNCTION
#define	_procedure			PROCEDURE
#define	_arguments			ARGUMENTS
#define	_funcbegin			FUNCBEGIN
#define	_funcend			FUNCEND
#define	_return				RESULT
#define	_finish				FINISH

#define	_if					IF
#define	_then				THEN
#define	_else				ELSE
#define	_for				FOR
#define	_do					DO
#define	_while				WHILE
#define	_break				BREAK
#define	_continue			CONTINUE
#define	_stop				STOP
#define	_action				ACTION
#define	_sizeof				SIZEOF

#define	_and				AND
#define	_or					OR
#define	_not				NOT
#define	_equal				EQUAL
#define	_different			DIFFERENT

#define	_const_integer		CONST_INTEGER
#define	_const_real			CONST_REAL
#define	_consrt_string		CONST_STRING
#define	_const_character	CONST_CHARACTER
#define	_const_boolean		CONST_BOOLEAN

#endif	// Header file end point.
