// VisualiserSuper.h
// Super-class for a FLIP execution visualizer.
// Virtual functions correspond to what has to be 
// implemented per specific visualizer.
// A. Savidis, deep re-factoring of first version from Y. Lilis.
// September, 2005.
// Project: FLIP Suite (Execution visualizer, Savidis, 2005).
//

#if	defined(HAS_VISUALIZER) && (defined(CONSOLE_VISUALIZER) ||defined(GRAPHICAL_VISUALIZER))

#ifndef	VISUALISERSUPER_H
#define	VISUALISERSUPER_H

#include <map>
#include <string>
#include <stdarg.h>
#include "Configurator.h"
#include "ExecutionEvents.h"

////////////////////////////////////////////////////////////////////
// Externalised by the Servant for memory access.
//
extern const char*			MemGetObjectStorageString (unsigned page, unsigned row);
extern const char*			MemGetObjectTypeString (unsigned page, unsigned row);
extern const char*			MemGetObjectName (unsigned page, unsigned row);
extern const std::string	MakeString (const char* format,...);
extern unsigned				MemMakeIndex (unsigned page, unsigned row);
extern int					GetARGC(void);
extern char**				GetARGV(void);

////////////////////////////////////////////////////////////////////

class VisualiserSuper {

	protected:
	std::map<std::string, bool>	notifyingEvents;	// Which execution events "notify".
	std::map<std::string, bool>	steppingEvents;		// Which execution events "step".

	// The following must be declared and defined by the derived
	// class as needed. Derived classes are mutually exclusive. The static
	// variables should never be explicitly initialized. The GetVisualiser()
	// function should create and assign upon the first call the derived visualiser.
	//
	static bool				cleanedUp;
	static VisualiserSuper* visualiser;

	VisualiserSuper (void)	{ Initialise(); visualiser = this; }
	virtual ~VisualiserSuper()	{ CleanUp(); visualiser = (VisualiserSuper*) 0; }

	public:

	static VisualiserSuper* GetVisualiser (void);

	/////////////////////////////////////////////////////////////////
	// Those are the only functions that the particular visualizer
	// MUST implement. All of them are pure virtual functions.
	//
	virtual void Message (const char* str) const = 0;	// An execution trace message.

	virtual void CellRefresh (
					const char*		name,	
					const char*		content,
					const char*		type, 
					const char*		storage, 
					const char*		operation, 
					unsigned		page, 
					unsigned		row
				) = 0;

	virtual void CellErase (
					unsigned	page, 
					unsigned	row
				) = 0;

	virtual void StepWait (void) const = 0;

	/////////////////////////////////////////////////////////////////
	// Refinement of those functions allows thead-enabled versions.
	//
	virtual bool IsNotifyingEvent (const std::string& name)
		{ return notifyingEvents[name]; }

	virtual void SetNotifyingEvent (const std::string& name, bool enabled)
		{ notifyingEvents[name] = enabled; }

	virtual bool IsSteppingEvent (const std::string& name)
		{ return steppingEvents[name]; }

	virtual void SetSteppingEvent (const std::string& name, bool enabled)
		{ steppingEvents[name] = enabled; }

	protected:

	/////////////////////////////////////////////////////////////////

#define	FUNCMACRO_NOTIFIER(x, declargs, callargs)			\
	static void On##x declargs {							\
		if (GetVisualiser())								\
			if(GetVisualiser()->IsNotifyingEvent(#x)) {		\
				GetVisualiser()->x callargs;				\
				if(GetVisualiser()->IsSteppingEvent(#x))	\
					GetVisualiser()->StepWait();			\
			}												\
	}

#define	FUNCMACRO_MESSAGER(x, declargs, callargs)	\
	void x declargs {								\
		Message(									\
			MakeString callargs.c_str()				\
		);											\
	}

#define	MSG(x)	Configurator::On##x

#define	FUNCMACRO_META(x, declargs, callargs)							\
	FUNCMACRO_MESSAGER(x, (declargs), (Configurator::On##x, callargs))	\
	FUNCMACRO_NOTIFIER(x, (declargs), (callargs))

	//************************

#define	LINE_DECLARGS	unsigned line
#define	LINE_CALLARGS	line

	FUNCMACRO_META(BlockEnter, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(BlockExit, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(IfEnter, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(IfExit, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(ForEnter, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(ForExit, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(WhileEnter, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(WhileExit, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(Break, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(Continue, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(CaseBegin, LINE_DECLARGS, LINE_CALLARGS)
	FUNCMACRO_META(CaseEnd, LINE_DECLARGS, LINE_CALLARGS)

	FUNCMACRO_META(ArrayAllocation, LINE_DECLARGS, LINE_CALLARGS)	// Used as length here

#define	STRING_BOOL_DECLARGS		const char* s, bool b
#define	STRING_BOOL_CALLARGS	s, b

	FUNCMACRO_META(ConditionValue, STRING_BOOL_DECLARGS, STRING_BOOL_CALLARGS)

	//************************

#define	STRING_UNSIGNED_DECLARGS	const char* func, unsigned line
#define	STRING_UNSIGNED_CALLARGS	func, line

	FUNCMACRO_META(FunctionEnter, STRING_UNSIGNED_DECLARGS, STRING_UNSIGNED_CALLARGS)
	FUNCMACRO_META(FunctionExit, STRING_UNSIGNED_DECLARGS, STRING_UNSIGNED_CALLARGS)
	FUNCMACRO_META(Finish, STRING_UNSIGNED_DECLARGS, STRING_UNSIGNED_CALLARGS)

	//************************

#define	UNSIGNED3_DECLARGS	unsigned page, unsigned row, unsigned line
#define	UNSIGNED3_CALLARGS	page, row, line

	FUNCMACRO_META(PageOf, UNSIGNED3_DECLARGS, UNSIGNED3_CALLARGS)
	FUNCMACRO_META(RowOf, UNSIGNED3_DECLARGS, UNSIGNED3_CALLARGS)
	FUNCMACRO_META(ContentOf, UNSIGNED3_DECLARGS, UNSIGNED3_CALLARGS)
	FUNCMACRO_META(PositionOf, UNSIGNED3_DECLARGS, UNSIGNED3_CALLARGS)	
	FUNCMACRO_META(ReleasePositions, UNSIGNED3_DECLARGS, UNSIGNED3_CALLARGS)
	FUNCMACRO_META(ArrayDestruction, UNSIGNED3_DECLARGS, UNSIGNED3_CALLARGS)
	FUNCMACRO_META(ArrayDeclaration, UNSIGNED3_DECLARGS, UNSIGNED3_CALLARGS)	// (startPage, startRow, length)

	//************************

#define	OP_DECLARGS				const char* op, unsigned p1, unsigned r1, unsigned p2, unsigned r2
#define	OP_NOTIFIER_CALLARGS	op, p1, r1, p2, r2
#define	OP_MESSAGER_CALLARGS	p1, r1, op, p2, r2

	FUNCMACRO_NOTIFIER(ArithmeticOperatorUsage, (OP_DECLARGS), (OP_NOTIFIER_CALLARGS))
	FUNCMACRO_MESSAGER(ArithmeticOperatorUsage, (OP_DECLARGS), (MSG(ArithmeticOperatorUsage), OP_MESSAGER_CALLARGS))
	FUNCMACRO_NOTIFIER(RelationalOperatorUsage, (OP_DECLARGS), (OP_NOTIFIER_CALLARGS))
	FUNCMACRO_MESSAGER(RelationalOperatorUsage, (OP_DECLARGS), (MSG(RelationalOperatorUsage), OP_MESSAGER_CALLARGS))

	//************************

#define	CHAR2_UNSIGNED_DECLARGS	const char* val, const char* func, unsigned line
#define	CHAR2_UNSIGNED_CALLARGS	val, func, line

	FUNCMACRO_META(Result, CHAR2_UNSIGNED_DECLARGS, CHAR2_UNSIGNED_CALLARGS)
	FUNCMACRO_META(FunctionUsage, CHAR2_UNSIGNED_DECLARGS, CHAR2_UNSIGNED_CALLARGS)

	//************************

#define	UNSIGNED4_DECLARGS	unsigned n1, unsigned n2, unsigned n3, unsigned n4
#define	UNSIGNED4_CALLARGS	n1, n2, n3, n4

	FUNCMACRO_META(RegionIndexing, UNSIGNED4_DECLARGS, UNSIGNED4_CALLARGS)		// (page, startRow, endRow, index)
	FUNCMACRO_META(RegionUsage, UNSIGNED4_DECLARGS, UNSIGNED4_CALLARGS)			// (page, startRow, endRow, line)
	FUNCMACRO_META(GrantPositions, UNSIGNED4_DECLARGS, UNSIGNED4_CALLARGS)		// (page, row, size, line)
	FUNCMACRO_META(ArrayIndexing, UNSIGNED4_DECLARGS, UNSIGNED4_CALLARGS)		// (startPage, startRow, length, index)

	//************************

#define	VOID_DECLARGS	(void)
#define	VOID_CALLARGS	()

	void RecordDeclaration (void)
		{ Message(Configurator::OnRecordDeclaration); }

	void RecordDestruction (void)
		{ Message(Configurator::OnRecordDestruction); }

	FUNCMACRO_NOTIFIER(RecordDeclaration, VOID_DECLARGS, VOID_CALLARGS)
	FUNCMACRO_NOTIFIER(RecordDestruction, VOID_DECLARGS, VOID_CALLARGS)

	//************************

#define	UNARYOP_DECLARGS	const char* op, unsigned page, unsigned row
#define	UNARYOP_CALLARGS	op, page, row

	FUNCMACRO_META(UnaryOperatorUsage, UNARYOP_DECLARGS, UNARYOP_CALLARGS)

	//************************

#define	VARDECL_DECLARGS				\
			const std::string&	name,	\
			const std::string&	value,	\
			const char*		type,		\
			const char*		storage,	\
			unsigned		page,		\
			unsigned		row

	void VarDeclaration (VARDECL_DECLARGS) {
		CellRefresh(
			name.c_str(), 
			value.c_str(), 
			type, 
			storage,
			"Variable declaration",
			page,
			row
		);
	}

#define	VARDECL_CALLARGS \
	name, value, type, storage, page, row

	FUNCMACRO_NOTIFIER(VarDeclaration, (VARDECL_DECLARGS), (VARDECL_CALLARGS))

	//************************

#define	VARWRITE_DECLARGS				\
			const std::string&	value,	\
			const char*		type,		\
			unsigned		page,		\
			unsigned		row

	void VarWrite (VARWRITE_DECLARGS) {
		CellRefresh(
			(char*) 0,	// i.e. don't change name
			value.c_str(), 
			type, 
			(char*) 0,	// i.e. don't change storage
			"Variable write",
			page,
			row
		);
	}

#define	VARWRITE_CALLARGS \
	value, type, page, row

	FUNCMACRO_NOTIFIER(VarWrite, (VARWRITE_DECLARGS), (VARWRITE_CALLARGS))

	//************************

#define	UNSIGNED2_DECLARGS	unsigned page, unsigned row
#define	UNSIGNED2_CALLARGS	page, row

	void VarDestruction (unsigned page, unsigned row) 
		{ CellErase(page, row); }

	FUNCMACRO_NOTIFIER(VarDestruction, (UNSIGNED2_DECLARGS), (UNSIGNED2_CALLARGS))

	/////////////////////////////////////////////////////////////////

#define INIT_FUNC(x)								\
			ExecutionEvents::SetOn##x(On##x);		\
			SetNotifyingEvent(#x, true);			\
			SetSteppingEvent(#x, false);

	void Initialise (void) {
		INIT_FUNC(BlockEnter)
		INIT_FUNC(BlockExit)
		INIT_FUNC(IfEnter)
		INIT_FUNC(IfExit)
		INIT_FUNC(ForEnter)
		INIT_FUNC(ForExit)
		INIT_FUNC(WhileEnter)
		INIT_FUNC(WhileExit)
		INIT_FUNC(Break)
		INIT_FUNC(Continue)
		INIT_FUNC(CaseBegin)
		INIT_FUNC(CaseEnd)
		INIT_FUNC(ConditionValue)
		INIT_FUNC(ArithmeticOperatorUsage)
		INIT_FUNC(RelationalOperatorUsage)
		INIT_FUNC(UnaryOperatorUsage)
		INIT_FUNC(ArrayDeclaration)
		INIT_FUNC(ArrayAllocation)
		INIT_FUNC(ArrayDestruction)
		INIT_FUNC(ArrayIndexing)
		INIT_FUNC(RecordDeclaration)
		INIT_FUNC(RecordDestruction)
		INIT_FUNC(FunctionEnter)
		INIT_FUNC(FunctionExit)
		INIT_FUNC(Result)
		INIT_FUNC(Finish)
		INIT_FUNC(PageOf)
		INIT_FUNC(RowOf)
		INIT_FUNC(ContentOf)
		INIT_FUNC(PositionOf)
		INIT_FUNC(RegionUsage)
		INIT_FUNC(RegionIndexing)
		INIT_FUNC(FunctionUsage)
		INIT_FUNC(GrantPositions)
		INIT_FUNC(ReleasePositions)
		INIT_FUNC(VarWrite)
		INIT_FUNC(VarDeclaration)
		INIT_FUNC(VarDestruction)
	}

	void CleanUp (void) {
		notifyingEvents.clear();
		steppingEvents.clear();
	}
};

//********************************

class VisualiserSuper_Init {
	private:
	static bool done;

	public:
	VisualiserSuper_Init (void);
};

static VisualiserSuper_Init visualiserInit;

#define	VISUALISER_INIT									\
	bool VisualiserSuper_Init::done;					\
	VisualiserSuper_Init::VisualiserSuper_Init (void) {	\
		if (!done) {									\
			VisualiserSuper::GetVisualiser();			\
			assert(VisualiserSuper::GetVisualiser());	\
		}												\
		else											\
			done = true;								\
	}

////////////////////////////////////////////////////////////////
// The linkage to mutually exclusive visualisers.
// They need declare the visualiser variable and implement
// the visualiser instance accessor and creation function. We
// never initialise explicitly, but rely on zeroing at start-up.
// We need only put the macro with the derived class name as a 
// parameter in the visualiser implementation file.
//
#define	VISUALISER_LINKAGE(_class)							\
bool				VisualiserSuper::cleanedUp;				\
VisualiserSuper*	VisualiserSuper::visualiser;			\
VisualiserSuper* VisualiserSuper::GetVisualiser (void) {	\
	/* After clean-up we disable visualisations. */			\
	if (!visualiser && !cleanedUp)	{						\
		new _class;											\
		assert(visualiser);									\
	}														\
	return visualiser;										\
}															\
VISUALISER_INIT

///////////////////////////////////////////////////////

#endif	// Header file end point.

#endif	// HAS_VISUALIZER
