// ConsoleVisualiser.cpp
// A console visualiser for the FLIP language.
// A. Savidis, October, 2005.
// Project: FLIP Execution Visualizer.
//

#if	defined(HAS_VISUALIZER) && defined(CONSOLE_VISUALIZER)

#include "ConsoleVisualiser.h"
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>

bool				ConsoleVisualiser::fileOutput;
FILE*				ConsoleVisualiser::file;

VISUALISER_LINKAGE(ConsoleVisualiser)

#define	CONSOLE_STDOUTPUT printf

////////////////////////////////////////////////////////////////
// Standard console I/O functions.
//
const char* ConsoleVisualiser::Input (const char* buff, unsigned len, FILE* fp) {
	return fgets(const_cast<char*>(buff), len, fp);
}

int ConsoleVisualiser::Output (const char* format, ...) {

	va_list args;
	va_start(args, format);

	int result;
	static char buffer[1024];
	result = vsprintf(buffer, format, args);

	if (fileOutput) {
		assert(file);
		fprintf(file, "%s", buffer);
	}
	else {
		printf("%s", buffer);
		fflush(stdout);
	}

	va_end(args);
	return result;
}

////////////////////////////////////////////////////////////////

void ConsoleVisualiser::CleanUp (void) {

	assert(!cleanedUp && visualiser);

	cleanedUp = true;
	delete ((ConsoleVisualiser*) visualiser);	// Derived class specific.
	visualiser = (VisualiserSuper*) 0;

	Output("Console execution tracer closed.\n");

	PostCleaner();
	exit(0);									// Clean-up exits immediately!
}

//*****************************

void ConsoleVisualiser::OnProgramStart (void) {
	CONSOLE_STDOUTPUT("Servant starts running the program.\n");
	Output("Console execution tracer activated.\n");
}

void ConsoleVisualiser::OnProgramEnd (void) {
	CONSOLE_STDOUTPUT("Servant finished the program normally.\n");
	CleanUp();
}

void ConsoleVisualiser::OnProgramStop (unsigned line) {
	CONSOLE_STDOUTPUT("Servant stopped execution at line %d.\n", line);
	CleanUp();
}

void ConsoleVisualiser::OnProgramError (const char* s) {
	CONSOLE_STDOUTPUT("Error: %s\nServant has interrupted program execution.\n", s);
}

//*****************************

ConsoleVisualiser::ConsoleVisualiser (void) {

	PreInitialiser();

	ExecutionEvents::SetInputFunction(Input);
	ExecutionEvents::SetOutputFunction(CONSOLE_STDOUTPUT);
	ExecutionEvents::SetCleanUpFunction(CleanUp);

	ExecutionEvents::SetOnProgramStart(OnProgramStart);
	ExecutionEvents::SetOnProgramEnd(OnProgramEnd);
	ExecutionEvents::SetOnProgramStop(OnProgramStop);
	ExecutionEvents::SetOnProgramError(OnProgramError);
}

ConsoleVisualiser::~ConsoleVisualiser() {
}

////////////////////////////////////////////////////////////////

void ConsoleVisualiser::Message (const char* str) const{
	Output("%s.\n", str);
}

//*****************************
// TODO: Need servant API to get name and storage of
// specific position.
//
void ConsoleVisualiser::CellRefresh (
		const char*		name,	
		const char*		content,
		const char*		type, 
		const char*		storage, 
		const char*		operation, 
		unsigned		page, 
		unsigned		row
	) {

	assert(visualiser);

	Message(
		MakeString(
			"%s: \"%s\":%s = %s:%s\n\t at position (%d, %d)", 
			operation,
			name ? name : MemGetObjectName(page, row),
			storage ? storage : MemGetObjectStorageString(page, row),
			content,
			type,
			page,
			row
		).c_str()
	);
}

//*****************************

void ConsoleVisualiser::CellErase (
		unsigned	page, 
		unsigned	row
	) {

	assert(visualiser);

	Message(
		MakeString(
			"Freeing: \"%s\":%s:%s\n\t from position (%d, %d)", 
			MemGetObjectName(page, row),
			MemGetObjectStorageString(page, row), 
			MemGetObjectTypeString(page, row),
			page,
			row
		).c_str()
	);
}

////////////////////////////////////////////////////////////////

#endif	// HAS_VISUALIZER && CONSOLE_VISUALIZER
