// GraphicalUI.cpp
// A GTK implementation of the Flip Visualizer
// Y. Lilis (only supervision and external API by A. Savidis)
// October, 2005.
// Project: FLIP Execution Visualizer, Savidis, 2005.
//
#if	defined(HAS_VISUALIZER) && defined(GRAPHICAL_VISUALIZER)

#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <glib.h>
#include <assert.h>
#include <string>
#include <string.h>
#include <errno.h>
#include "GraphicalUILinkage.h"
#include "NotebookMirror.h"

//////////////////////////////////
// Visualizer GUI specific values
//
#define WINDOW_WIDTH		800
#define WINDOW_HEIGHT		600

#define NOTEBOOK_COLUMNS	4	//name, value, type, storage

///////////////////////////////////////////////////////////////////////////
// These are the same definitions with the Servant PAGE_ROWS and MAX_PAGES
//
#define	PAGE_ROWS	60
#define	MAX_PAGES	300

#define IF_DESTROYED_EXIT()	do { if(!window) return; } while(false)

//////////////////////////////////////
// Function externalized from Servant
//
extern const std::string MakeString(const char *format, ...);

////////////////////////////////////////
// Functions found in this file. These functions 
// manage the Gtk Implementation of the UI. They
// are for private use and declared static.
//

static void OnPreviousClicked (GtkButton *button, gpointer user_data);	//Called upon pressing "previous"

static void OnNextClicked (GtkButton *button, gpointer user_data);		//Called upon pressing "next"

static void OnGotoClicked (GtkButton *button, gpointer user_data);		//Called upon pressing "goto"

static void OnStepClicked (GtkButton *button, gpointer user_data);		//Called upon pressing "step"

static void OnInputActivate (GtkEntry *entry, gpointer user_data);		//Called upon pressing return on Input entry

//////////////////////////////////////////////
// The Menu Bar implementation and callbacks
//
namespace GTKMENU {
static void Save (const char* name);									//Saves the whole program state to a file
static void StoreFilename (GtkWidget* widget, gpointer user_data);		//Gets the filename inserted in the appropriate dialog box
static void CreateFileSelection (void);									//

static void SetNotifyingState (GtkAction *action);						//Called when changing an notifying event
static void SetSteppingState (GtkAction *action);						//Called when changing a steping event

static void File (GtkAction* action);									//Menu->File event dispatcher
static void ActivateEmail (GtkAboutDialog* about, const gchar* link, gpointer data);	//Called upon clicking an author or documentor name
static void About (GtkAction* action, GtkWidget* window);				//Shows information about the Visualizer

static void Create (GtkWidget* window, GtkWidget* content);				//Creates the Menu Bar for the Window of the UI
}

//****************************

static void RemoveNewLines (char* s);									//Helper string manipulation function

static void GtkEnd (void);												//Ends the Visualizer main loop

static void GtkStatus (const char* str);								//Sets the program Status Bar

static void* GtkMain (void *unused);									//Visualizer thread that handles pending GTK events

static void GtkSetCurrentPage (unsigned page);							//Sets the Visualizer notebook to this page

static const char* GtkGetTextViewText (GtkTextView* textview);			//GetTextView getText wrapper function

static GtkWidget* GtkCreateWindow (void);								//Creates the window for the UI

///////////////////////////////////////////////
// TODO: Add this if we want to have the title
//
//void GtkSetTitle (void);										//Sets the program name as the Title.


//////////////////////////////////////////////////
// GUI data needed globally. All of them are for
// private use and therefore declared static.
//
static GtkWidget*	notebook[NOTEBOOK_COLUMNS + 1][PAGE_ROWS + 1];
static GtkWidget*	TraceText;
static GtkWidget*	OutputText;
static GtkWidget*	InputEntry;
static GtkWidget*	statusbar;
static GtkWidget*	window;

static bool	GtkDone = false;

static bool messageAlreadyShown = false;

static unsigned	CurrentPage = 0;

static char*	input = (char *) 0;
static bool		inputDone = false;

static const char*	SelectedFilename = (const char*) 0;

///////////////////////////////////////////////////////////////////
// Mutexes and conditions for mutual exclusion and synchronization
//
static GMutex*	step_mutex;
static GMutex*	wait_mutex;
static GMutex*	visual_mutex;
static GCond*	step_cond;
static GCond*	wait_cond;
static GCond*	visual_cond;

///////////////////////////////////////////
// The argments passed by the main thread
//
static int _global_argc;
static char **_global_argv;


//////////////////////////////////////////////////
// GRAPHICALUI namespace implementation functions
//


////////////////////////////////////////////
// The initialization and cleanup of the UI
//

void GRAPHICALUI::Create (int argc, char** argv) {
	
	//////////////////////////////////////////////////////////
	//TODO: Get the argc and argv after main has been called,
	// as the values they have here is not the proper one.

	_global_argc = argc;
	_global_argv = argv;

	GtkDone = false;
	if(!g_thread_supported ()) {
		g_thread_init (NULL);
		gdk_threads_init();
	}
	
	g_thread_create(GtkMain, NULL, 1, NULL);

	step_mutex = g_mutex_new();
	wait_mutex = g_mutex_new();
	visual_mutex = g_mutex_new();
	step_cond = g_cond_new();
	wait_cond = g_cond_new();
	visual_cond = g_cond_new();

	g_mutex_lock(wait_mutex);
	g_cond_wait(wait_cond, wait_mutex);
	g_mutex_unlock(wait_mutex);
}


void GRAPHICALUI::Destroy (void){ 
  IF_DESTROYED_EXIT();
  GtkEnd(); 
}


///////////////////////////////////////////////////////////////
// Functions with which the main thread  interacta with the UI.
//

void GRAPHICALUI::Refresh (unsigned page, unsigned row) {
	IF_DESTROYED_EXIT();
  
	// Not current page so don't change anything at screen
	//
	if(page != CurrentPage)	
		return;

	std::string name = "", value = "", type = "", storage = "";
	
	NotebookMirror::GetSingleton()->GrantCellAccess();
	if(NotebookMirror::GetSingleton()->IsCellOccupied(page, row))
		NotebookMirror::GetSingleton()->GetCellContent(page, row, name, value, type, storage);
	NotebookMirror::GetSingleton()->ReleaseCellAccess();
	
	for(unsigned i = 1; i <= NOTEBOOK_COLUMNS; ++i) {
		const char *text = (const char *) 0;

		switch(i)
		{
			case 1:	text = name.c_str();	break;
			case 2:	text = value.c_str();	break;
			case 3:	text = type.c_str();	break;
			case 4: text = storage.c_str();	break;
			default:	assert(false);
		}
		RemoveNewLines((char *) text);
		
		if(!GtkDone)
			gtk_entry_set_text((GtkEntry *)notebook[i][row + 1], text);
	}
}

//****************************

void GRAPHICALUI::Finished (void) {
	IF_DESTROYED_EXIT();
	GtkStatus("Program finished: press 'Step' to exit");
	messageAlreadyShown = true;
}

void GRAPHICALUI::Started (void) {
	IF_DESTROYED_EXIT();
	GtkStatus("Program started: press 'Step' to begin");
	messageAlreadyShown = true;
}

void GRAPHICALUI::StepWait (void) {
	IF_DESTROYED_EXIT();
	
	if(!messageAlreadyShown)
		GtkStatus("Stepping into event. Press 'Step' to continue");
	else
		messageAlreadyShown = false;
		
	g_mutex_lock(step_mutex);
	g_cond_wait(step_cond, step_mutex);
	g_mutex_unlock(step_mutex);
	GtkStatus("");
}

//****************************

void GRAPHICALUI::InputWait (char* buffer, unsigned len) {

	IF_DESTROYED_EXIT();
  
	inputDone = false;
	GtkStatus("Waiting for input");
	g_mutex_lock(wait_mutex);
	g_cond_wait(wait_cond, wait_mutex);
	g_mutex_unlock(wait_mutex);
	GtkStatus("");
	
	if(input) {
		if(strlen(input) >= len)
			input[len - 1] = '\0';
		sprintf(buffer, "%s", input);
	}
	else
		*buffer = '\0';

	//for the echo
	//
	if(*buffer)
		GRAPHICALUI::Output(MakeString("%s\n", buffer).c_str());

	free(input);
	input = (char *) 0;
}

//****************************

void GRAPHICALUI::Output (const char* msg) {
	IF_DESTROYED_EXIT();
	
	g_mutex_lock(visual_mutex);
	g_cond_wait(visual_cond, visual_mutex);
		if(!GtkDone) {
			GtkTextBuffer *buf = gtk_text_view_get_buffer((GtkTextView *)OutputText);
			GtkTextIter iter;

			gtk_text_buffer_get_end_iter(buf, &iter);
			gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(OutputText), &iter, 0, 1, 0, 0);
			gtk_text_buffer_insert(buf, &iter, msg, -1);
		}
	g_mutex_unlock(visual_mutex);
}

//****************************

void GRAPHICALUI::Trace (const char* msg) {
	IF_DESTROYED_EXIT();
  
	static char str[1024];
	sprintf(str, "%s.\n", msg);
	g_mutex_lock(visual_mutex);
	g_cond_wait(visual_cond, visual_mutex);
		if(!GtkDone) {
			GtkTextBuffer *buf = gtk_text_view_get_buffer((GtkTextView *)TraceText);
			GtkTextIter iter;

			gtk_text_buffer_get_end_iter(buf, &iter);
			gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(TraceText), &iter, 0, 1, 0, 0);
			gtk_text_buffer_insert(buf, &iter, str, -1);
		}
	g_mutex_unlock(visual_mutex);
}


////////////////////////////////////////////////////
// Mutex wrapper function for using the GTK mutexes
//

unsigned GRAPHICALUI::CreateMutex (void)
	{ return (unsigned) g_mutex_new(); }

void GRAPHICALUI::DestroyMutex (unsigned id)
	{ g_mutex_free((GMutex *) id); }

void GRAPHICALUI::GrantMutex (unsigned id)
	{ g_mutex_lock((GMutex *) id); }

void GRAPHICALUI::ReleaseMutex (unsigned id)
	{ g_mutex_unlock((GMutex *) id); }


//********************************************************************

//////////////////////////////////////////////////////////////////////
// The functions that follow are part of the UI implementation. They
// are functions for our private use and therefore declared static.
//

//********************************************************************

////////////////////////////
// GUI callback functions
//

static void OnPreviousClicked (GtkButton *button, gpointer user_data) {
	if(CurrentPage)
		GtkSetCurrentPage(--CurrentPage);
}

//****************************

static void OnNextClicked (GtkButton *button, gpointer user_data) {
	if(CurrentPage != MAX_PAGES - 1)
		GtkSetCurrentPage(++CurrentPage);
}

//****************************

static void OnGotoClicked (GtkButton *button, gpointer user_data) {
	char* text= gtk_combo_box_get_active_text(GTK_COMBO_BOX(user_data));
	unsigned page;
	if(text && (page = atoi(text)) >= 0 && page < MAX_PAGES)
		GtkSetCurrentPage(CurrentPage = page);
	else
		gtk_combo_box_set_active(GTK_COMBO_BOX(user_data), 0);
}

//****************************

static void OnStepClicked (GtkButton *button, gpointer user_data) {
	g_mutex_lock(step_mutex);
	g_cond_signal(step_cond);
	g_mutex_unlock(step_mutex);
}

//****************************

static void OnInputActivate (GtkEntry *entry, gpointer user_data) {
	g_mutex_lock(wait_mutex);
	input = strdup((char *)gtk_entry_get_text(entry));
	gtk_entry_set_text(entry, "");
	inputDone = true;
	g_cond_signal(wait_cond);
	g_mutex_unlock(wait_mutex);
}

/////////////////////////////////////
// GTKMENU namespace implementation
//

static void GTKMENU::SetNotifyingState (GtkAction *action) {	
	const char *name = gtk_action_get_name (action);
	EVENTAPI::SetNotifyingEvent(
		name, 
		!EVENTAPI::IsNotifyingEvent(name)
	);
}


static void GTKMENU::SetSteppingState (GtkAction *action) {	
	const char *name = gtk_action_get_name (action);
	assert(!strcmp(name + strlen(name) - strlen("Step"), "Step"));	//must end with "Step"

	std::string s = name;
	s = s.substr(0, strlen(name) - strlen("Step"));	//cut the trailing "Step"
	EVENTAPI::SetSteppingEvent(
		s,
		!EVENTAPI::IsSteppingEvent(s)
	);
}

/////////////////////////////////////////////
// Functions for saving the execution TraceText
//

static void GTKMENU::Save (const char * name) {
	char table[1024];
	FILE *fp = fopen(name, "w");

	if(fp) {
		fputs("FLIP VISUALIZER : ", fp);
		if(_global_argv && *_global_argv)
			fputs(*_global_argv, fp);

		fputs("\n\n\tOUTPUT\n", fp);
		fputs(GtkGetTextViewText(GTK_TEXT_VIEW(OutputText)), fp);
		fputs("\n\n\tGLOBAL MESSAGES\n", fp);
		fputs(GtkGetTextViewText(GTK_TEXT_VIEW(TraceText)), fp);

		fputs("\n\n\tNOTEBOOK\n\nROW\t\tNAME\tVALUE\tTYPE\tSTORAGE TYPE\n\n", fp);

		unsigned page, row;
		for(page = 0; page < MAX_PAGES; ++page) {
			sprintf(table, "\nPAGE %d\n", page);
			fputs(table, fp);
			for(row = 0; row < PAGE_ROWS; ++row) {
				std::string name = "", value = "", type = "", storage = "";
				bool occupied = false;
	
				NotebookMirror::GetSingleton()->GrantCellAccess();
				if(NotebookMirror::GetSingleton()->IsCellOccupied(page, row)) {
					occupied = true;
					NotebookMirror::GetSingleton()->GetCellContent(page, row, name, value, type, storage);
				}
				NotebookMirror::GetSingleton()->ReleaseCellAccess();
	
				if(occupied)
					sprintf(table, 
							"ROW %d\t:\t%s\t%s\t%s\t%s\n", 
							row, 
							name.c_str(), 
							value.c_str(), 
							type.c_str(), 
							storage.c_str()
					);
				else
					//sprintf(buf, "ROW %d\t:\n", row);	// this for full
					*table = '\0';					// this only for occupied ones

				fputs(table, fp);
			}
		}			
		fclose(fp);
	}
	else {
		GtkWidget *dialog;
		dialog = gtk_message_dialog_new (
			GTK_WINDOW(window),
            GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_ERROR,
            GTK_BUTTONS_CLOSE,
            "Error creating file '%s': %s",
            name, 
			g_strerror (errno)
		);
		gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy (dialog);
	}
}

//****************************

static void GTKMENU::StoreFilename (GtkWidget *widget, gpointer user_data) {
   GtkWidget *file_selector = GTK_WIDGET (user_data);
   SelectedFilename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_selector));
   
   Save(SelectedFilename);
}

//****************************

static void GTKMENU::CreateFileSelection (void) {
	GtkWidget *file_selector = gtk_file_selection_new ("Save as");

	g_signal_connect (
		GTK_FILE_SELECTION (file_selector)->ok_button,
		 "clicked",
		 G_CALLBACK (StoreFilename),
		 file_selector
	);

	g_signal_connect_swapped (
		GTK_FILE_SELECTION (file_selector)->ok_button,
		"clicked",
		G_CALLBACK (gtk_widget_destroy), 
		file_selector
	);

	g_signal_connect_swapped (
		GTK_FILE_SELECTION (file_selector)->cancel_button,
		"clicked",
		G_CALLBACK (gtk_widget_destroy),
		file_selector
	); 

	gtk_widget_show (file_selector);
}

////////////////////////////////////////////////
// Functions for handling the Menu bar actions
//

static void GTKMENU::File (GtkAction *action) {
	const char *name = gtk_action_get_name(action);

	if(!strcmp(name, "Save")) {
		if(SelectedFilename)
			Save(SelectedFilename);
		else
			CreateFileSelection();
	}
	else if(!strcmp(name, "SaveAs"))
		CreateFileSelection();
	else if(!strcmp(name, "Quit"))
		GtkDone = true;
	else
		assert(false);
}

//****************************

static void GTKMENU::ActivateEmail (GtkAboutDialog *about, const gchar *link, gpointer data) {
#ifndef WIN32
	char buf[1024];
	printf("Write your message and press CTRL+D to exit (noted by ^D) :\n");
	sprintf(buf, "mail %s", link);
	system(buf);
#else
	g_print("Unable to send mail to %s\n", link);
#endif
}

//****************************

static void GTKMENU::About (GtkAction *action, GtkWidget *window) {
	const gchar *author[] = {
		"HCI Laboratory, FORTH-ICS, 2005",
		"Concept, architecture and programming : Anthony Savidis",
		"E-mail: <as@ics.forth.gr>",
		"Co-programming and UI: John Lilis",
		NULL
	};

	const gchar *documentors[] = {
		"User guide: John Lilis",
		"Further contact: Anthony Savidis, FORTH-ICS",
		"<as@ics.forth.gr>",
		NULL
	};


	const gchar *license =
		"This program is a research outcome of HCI laboratory\n"
		"at FORTH-ICS. Currently it is provided for educational\n"
		"purposes only, at the Department of Computer Science,\n"
		"University of Crete. To gain a free deployment license\n"
		"please contact Anthony Savidis, as@ics.forth.gr";
	
	gtk_about_dialog_set_email_hook (ActivateEmail, NULL, NULL);
	gtk_show_about_dialog (GTK_WINDOW (window),
							"name", "Flip Visualizer",
							"version", "1.0",
							"copyright", "Anthony Savidis. HCI Lab, FORTH-ICS, 2005.",
							"license", license,
							"comments", "Execution visualizer for the Flip language.",
							"authors", author,
							"documenters", documentors,
							NULL);
}

//****************************

static void GTKMENU::Create (GtkWidget *window, GtkWidget *content) {
	static GtkActionEntry entries[] = {
		{ "FileMenu", NULL, "_File" },
		{ "VisualizedActive", NULL, "Visualized _Events" },
		{ "VisualizedStep", NULL, "Visualized _Step" },
		{ "HelpMenu", NULL, "_Help" },
		//name, stock id label
		
		{ "Save", GTK_STOCK_SAVE, "_Save","<control>S",  
			"Save current file", G_CALLBACK (File) },
		{ "SaveAs", GTK_STOCK_SAVE,	"Save _As...", NULL,     
			"Save to a file", G_CALLBACK (File) },
		{ "Quit", GTK_STOCK_QUIT, "_Quit", "<control>Q",     
			"Quit",	G_CALLBACK (File) },
		{ "About", NULL, "_About", "<control>A",     
			"About", G_CALLBACK (About) },
		//name, stock id, label, accelerator, tooltip
	};

	static guint n_entries = G_N_ELEMENTS (entries);

#define TOGGLE_ENTRIES(name, x, func, value)	\
	static GtkToggleActionEntry name[] = {\
		{ "BlockEnter" x, NULL, "Block enter", NULL, "Block enter", G_CALLBACK (func), value },\
		{ "BlockExit" x, NULL, "Block exit", NULL, "Block exit", G_CALLBACK (func), value },\
	\
		{ "IfEnter" x, NULL, "If enter", NULL, "If enter", G_CALLBACK (func), value },\
		{ "IfExit" x, NULL, "If exit", NULL, "If exit", G_CALLBACK (func), value },\
	\
		{ "ForEnter" x, NULL, "For enter", NULL, "For enter", G_CALLBACK (func), value },\
		{ "ForExit" x, NULL, "For exit", NULL, "For exit", G_CALLBACK (func), value },\
	\
		{ "WhileEnter" x, NULL, "While enter", NULL, "While enter", G_CALLBACK (func), value },\
		{ "WhileExit" x, NULL, "While exit", NULL, "While exit", G_CALLBACK (func), value },\
	\
		{ "CaseEnter" x, NULL, "Case enter", NULL, "Case enter", G_CALLBACK (func), value },\
		{ "CaseExit" x, NULL, "Case exit", NULL, "Case exit", G_CALLBACK (func), value },\
	\
		{ "Break" x, NULL, "Break", NULL, "Break", G_CALLBACK (func), value },\
		{ "Continue" x, NULL, "Continue", NULL, "Continue", G_CALLBACK (func), value },\
	\
		{ "ArithmeticOperatorUsage" x, NULL, "Arithmetic operator", NULL, \
			"Using arithmetic operator" x, G_CALLBACK (func), value },\
		{ "RelationalOperatorUsage" x, NULL, "Relational operator", NULL, \
			"Using relational operator" x, G_CALLBACK (func), value },\
		{ "UnaryOperatorUsage" x, NULL, "Unary operator", NULL, \
			"Using unary operator" x, G_CALLBACK (func), value },\
	\
		{ "VarWrite" x, NULL, "Variable writing", NULL, "Writing variable", G_CALLBACK (func), value },\
		{ "VarDeclaration" x, NULL, "Variable declaration", NULL, \
			"Declaring variable" x, G_CALLBACK (func), value },\
		{ "VarDestruction" x, NULL, "Variable destruction", NULL, \
			"Destroying variable" x, G_CALLBACK (func), value },\
	\
		{ "ArrayDeclaration" x, NULL, "Array declaration", NULL, \
			"Declaring array" x, G_CALLBACK (func), value },\
		{ "ArrayAllocation" x, NULL, "Array alocation", NULL, \
			"Allocating array" x, G_CALLBACK (func), value },\
		{ "ArrayDestruction" x, NULL, "Array destruction", NULL, \
			"Destroying array" x, G_CALLBACK (func), value },\
		{ "ArrayIndexing" x, NULL, "Array indexing", NULL, \
			"Indexing array" x, G_CALLBACK (func), value },\
	\
		{ "FunctionEnter" x, NULL, "Function enter", NULL, "Function enter", G_CALLBACK (func), value },\
		{ "FunctionExit" x, NULL, "Function exit", NULL, "Function exit", G_CALLBACK (func), value },\
	\
		{ "Result" x, NULL, "Function result", NULL, "Function result", G_CALLBACK (func), value },\
		{ "Finish" x, NULL, "Function finish", NULL, "Function finish", G_CALLBACK (func), value },\
	\
		{ "PageOf" x, NULL, "PageOf", NULL, "PageOf", G_CALLBACK (func), value },\
		{ "RowOf" x, NULL, "RowOf", NULL, "RowOf", G_CALLBACK (func), value },\
		{ "ContentOf" x, NULL, "ContentOf", NULL, "ContentOf", G_CALLBACK (func), value },\
		{ "PositionOf" x, NULL, "PositionOf", NULL, "PositionOf", G_CALLBACK (func), value },\
		{ "GrantPositions" x, NULL, "GrantPositions", NULL, "GrantPositions", G_CALLBACK (func), value },\
		{ "ReleasePositions" x, NULL, "ReleasePositions", NULL, "ReleasePositions", G_CALLBACK (func), value },\
	\
		{ "RegionUsage" x, NULL, "UseRegion", NULL, "Using region", G_CALLBACK (func), value },\
		{ "RegionIndexing" x, NULL, "Region indexing", NULL, "Indexing region", G_CALLBACK (func), value },\
	\
		{ "FunctionUsage" x, NULL, "Library function", NULL, "Using library function", G_CALLBACK (func), value },\
	\
		{ "RecordDeclaration" x, NULL, "Record declaration", NULL, \
			"Declaring record" x, G_CALLBACK (func), value },\
		{ "RecordDestruction" x, NULL, "Record destruction", NULL, \
			"Destroying record" x, G_CALLBACK (func), value },\
	};

	TOGGLE_ENTRIES(toggle_entries_active, "", GTKMENU::SetNotifyingState, TRUE)
	TOGGLE_ENTRIES(toggle_entries_step, "Step", GTKMENU::SetSteppingState, FALSE)

	static guint n_toggle_entries_active = G_N_ELEMENTS (toggle_entries_active);
	static guint n_toggle_entries_step = G_N_ELEMENTS (toggle_entries_step);

	static const char *ui_info = 
	"<ui>"
	"  <menubar name='MenuBar'>"
	"    <menu action='FileMenu'>"
	"      <menuitem action='Save'/>"
	"      <menuitem action='SaveAs'/>"
	"      <separator/>"
	"      <menuitem action='Quit'/>"
	"    </menu>"
	
	"	<menu name='Active' action='VisualizedActive'>"
	"      <menuitem action='BlockEnter'/>"
	"      <menuitem action='BlockExit'/>"
	"      <menuitem action='IfEnter'/>"
	"      <menuitem action='IfExit'/>"
	"      <menuitem action='ForEnter'/>"
	"      <menuitem action='ForExit'/>"
	"      <menuitem action='WhileEnter'/>"
	"      <menuitem action='WhileExit'/>"
	"      <menuitem action='Break'/>"
	"      <menuitem action='Continue'/>"
	"      <menuitem action='CaseEnter'/>"
	"      <menuitem action='CaseExit'/>"
	"      <separator/>"
	"      <menuitem action='ArithmeticOperatorUsage'/>"
	"      <menuitem action='RelationalOperatorUsage'/>"	
	"      <menuitem action='UnaryOperatorUsage'/>"
	"      <separator/>"
	// These events must always be visualized, so no configuration.
/*	"      <menuitem action='VarWrite'/>"
	"      <menuitem action='VarDeclaration'/>"
	"      <menuitem action='VarDestruction'/>"	*/
	"      <menuitem action='ArrayDeclaration'/>"
	"      <menuitem action='ArrayAllocation'/>"
	"      <menuitem action='ArrayDestruction'/>"
	"      <menuitem action='ArrayIndexing'/>"
	"      <separator/>"
	"      <menuitem action='FunctionEnter'/>"
	"      <menuitem action='FunctionExit'/>"
	"      <menuitem action='Result'/>"
	"      <menuitem action='Finish'/>"
	"      <separator/>"
	"      <menuitem action='RecordDeclaration'/>"
	"      <menuitem action='RecordDestruction'/>"
	"      <menuitem action='RegionUsage'/>"
	"      <menuitem action='RegionIndexing'/>"
	"      <separator/>"
	"      <menuitem action='PageOf'/>"
	"      <menuitem action='RowOf'/>"
	"      <menuitem action='ContentOf'/>"
	"      <menuitem action='PositionOf'/>"
	"      <menuitem action='GrantPositions'/>"
	"      <menuitem action='ReleasePositions'/>"
	"      <menuitem action='FunctionUsage'/>"
	"    </menu>"

	"	<menu name='Step' action='VisualizedStep'>"
	"      <menuitem action='BlockEnterStep'/>"
	"      <menuitem action='BlockExitStep'/>"
	"      <menuitem action='IfEnterStep'/>"
	"      <menuitem action='IfExitStep'/>"
	"      <menuitem action='ForEnterStep'/>"
	"      <menuitem action='ForExitStep'/>"
	"      <menuitem action='WhileEnterStep'/>"
	"      <menuitem action='WhileExitStep'/>"
	"      <separator/>"
	"      <menuitem action='ArithmeticOperatorUsageStep'/>"
	"      <menuitem action='RelationalOperatorUsageStep'/>"	
	"      <menuitem action='UnaryOperatorUsageStep'/>"
	"      <separator/>"
	"      <menuitem action='ArrayDeclarationStep'/>"
	"      <menuitem action='ArrayAllocationStep'/>"
	"      <menuitem action='ArrayDestructionStep'/>"
	"      <menuitem action='ArrayIndexingStep'/>"
	"      <separator/>"
	"      <menuitem action='FunctionEnterStep'/>"
	"      <menuitem action='FunctionExitStep'/>"
	"      <menuitem action='ResultStep'/>"
	"      <menuitem action='FinishStep'/>"
	"      <separator/>"
	"      <menuitem action='RecordDeclarationStep'/>"
	"      <menuitem action='RecordDestructionStep'/>"
	"      <menuitem action='RegionUsageStep'/>"
	"      <menuitem action='RegionIndexingStep'/>"
	"      <separator/>"
	"      <menuitem action='PageOfStep'/>"
	"      <menuitem action='RowOfStep'/>"
	"      <menuitem action='ContentOfStep'/>"
	"      <menuitem action='PositionOfStep'/>"
	"      <menuitem action='GrantPositionsStep'/>"
	"      <menuitem action='ReleasePositionsStep'/>"

	"      <menuitem action='FunctionUsageStep'/>"
	"      <separator/>"
	"      <menuitem action='VarWriteStep'/>"
	"      <menuitem action='VarDeclarationStep'/>"
	"      <menuitem action='VarDestructionStep'/>"
	"    </menu>"

	"    <menu action='HelpMenu'>"
	"      <menuitem action='About'/>"
	"    </menu>"
	"  </menubar>"
	"</ui>";

	GtkWidget *table;
	GtkWidget *bar;
	GtkActionGroup *action_group;
	GtkUIManager *merge;
	GError *error = NULL;

	table = gtk_table_new (1, 3, FALSE);
	gtk_container_add (GTK_CONTAINER (window), table);

	action_group = gtk_action_group_new ("AppWindowActions");
	gtk_action_group_add_actions (action_group, entries, n_entries, window);
	gtk_action_group_add_toggle_actions (action_group, toggle_entries_active, 
											n_toggle_entries_active, NULL);
	gtk_action_group_add_toggle_actions (action_group, toggle_entries_step, 
											n_toggle_entries_step, NULL);

	merge = gtk_ui_manager_new ();
	g_object_set_data_full (G_OBJECT (window), "ui-manager", merge, g_object_unref);
	gtk_ui_manager_insert_action_group (merge, action_group, 0);
	gtk_window_add_accel_group (GTK_WINDOW (window), 
		gtk_ui_manager_get_accel_group (merge));

	gtk_ui_manager_set_add_tearoffs (merge, TRUE);	

	if (!gtk_ui_manager_add_ui_from_string (merge, ui_info, -1, &error)) {
		g_message ("Error building menus : %s", error->message);
		g_error_free (error);
		exit(1);
	}

	bar = gtk_ui_manager_get_widget (merge, "/MenuBar");
	gtk_widget_show (bar);
	gtk_table_attach (GTK_TABLE (table), bar, 0, 1, 0, 1,
		(GtkAttachOptions)(GTK_EXPAND | GTK_FILL), 
		(GtkAttachOptions)(0), 0, 0);

	gtk_table_attach (GTK_TABLE (table), content, 0, 1, 2, 3,
		(GtkAttachOptions)(GTK_EXPAND | GTK_FILL), 
		(GtkAttachOptions)(GTK_EXPAND | GTK_FILL), 0, 0);

	if (!GTK_WIDGET_VISIBLE (window))
		gtk_widget_show_all (window);
	else {
		gtk_widget_destroy (window);
		window = NULL;
	}
}

//////////////////////////////////////////////
// GUI implementation helper & core functions.
//

static void RemoveNewLines (char *s) {
	while(*s) {
		if(*s=='\n')
			*s = ' ';
		++s;
	}
}

//****************************

static void GtkEnd (void) {
	GtkDone = true;
}

//****************************

static void GtkStatus (const char *str) {
	if(!GtkDone) {
		g_mutex_lock(visual_mutex);
		g_cond_wait(visual_cond, visual_mutex);
		gtk_statusbar_push(GTK_STATUSBAR(statusbar), 0, str);
		g_mutex_unlock(visual_mutex);
	}
}

//****************************

//////////////////////////////////////////////////////
// TODO: Export this function (or a wrapper one) to 
//the main thread for setting the program title.

/*
static void GtkSetTitle (void) {
	if(window && _global_argv) {
		char temp[1024];
		sprintf(temp, "JET VISUALIZER - %s", *_global_argv);
		gtk_window_set_title (GTK_WINDOW (window), temp);
	}
}*/

//****************************

static void* GtkMain (void *unused) {

    gtk_set_locale();
    gtk_init(&_global_argc, &_global_argv);
	
    window = GtkCreateWindow();
	gtk_widget_show (window);

	g_signal_connect ((gpointer) window, "destroy", G_CALLBACK(GtkEnd), NULL);

	g_mutex_lock(wait_mutex);
	g_cond_signal(wait_cond);	//signal the main thread to continue
	g_mutex_unlock(wait_mutex);

	//visual loop
	while(!GtkDone) {
		gdk_threads_enter();
		while(gtk_events_pending() && !GtkDone) {
			g_mutex_lock(visual_mutex);
			gtk_main_iteration();
			g_cond_signal(visual_cond);
			g_mutex_unlock(visual_mutex);
		}
		gdk_threads_leave();
	}

	window = (GtkWidget *) 0;
	
	g_cond_signal(visual_cond);
	g_cond_signal(step_cond);
	g_cond_signal(wait_cond);
	/*
	g_mutex_free(step_mutex);
	g_mutex_free(wait_mutex);
	g_mutex_free(visual_mutex);
	g_cond_free(step_cond);
	g_cond_free(wait_cond);
	g_cond_free(visual_cond);
	*/
	gtk_exit(0);
	exit(0);
    return (void *) 0;
}

//****************************

static void GtkSetCurrentPage(unsigned page) {
	char temp[50];
	
	for(unsigned j = 1; j <= PAGE_ROWS; ++j) {
		sprintf(temp, " (%d, %d) ", page, j - 1);
		gtk_label_set_text(GTK_LABEL(notebook[0][j]), temp);
	}

	for(unsigned row = 0; row < PAGE_ROWS; ++row)
		GRAPHICALUI::Refresh(page, row);
}

//****************************

static const char *GtkGetTextViewText (GtkTextView *textview) {
	GtkTextBuffer *buf = gtk_text_view_get_buffer((GtkTextView *)textview);
	GtkTextIter start_iter, end_iter;
	
	gtk_text_buffer_get_start_iter(buf, &start_iter);
	gtk_text_buffer_get_end_iter(buf, &end_iter);

	return gtk_text_buffer_get_text(buf, &start_iter, &end_iter, FALSE);
}

////////////////////////////////////////
// The actual window creating function.
// Most of it is auto-generate by glade.
//

static GtkWidget* GtkCreateWindow (void) {
	unsigned i,j;
	char temp[1024];

	GtkWidget *window1;
	GtkWidget *table1;
	GtkWidget *hbox1;
	GtkWidget *hpaned;
	GtkWidget *vbox8;
	GtkWidget *label8;
	GtkWidget *scrolledwindow5;
	GtkWidget *viewport1;
	GtkWidget *table3;
	GtkWidget *vbox9;
	GtkWidget *button_previous;
	GtkWidget *button_next;
	GtkWidget *button_step;
	GtkWidget *button_goto;
	GtkWidget *hbox2;
	GtkWidget *vbox6;
	GtkWidget *label5;
	GtkWidget *label6;
	GtkWidget *label7;
	GtkWidget *scrolledwindow1;
	GtkWidget *vbox1;
	GtkWidget *scrolledwindow2;
	GtkWidget *combo;
	GtkWidget *frame;
	GtkWidget *box;

	window1 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (window1), "VISUALIZER");
	gtk_window_set_default_size(GTK_WINDOW(window1), WINDOW_WIDTH, WINDOW_HEIGHT);

	table1 = gtk_table_new (2, 1, FALSE);
	gtk_widget_show (table1);

	GTKMENU::Create(window1, table1);
	
	statusbar = gtk_statusbar_new ();
	gtk_table_attach (GTK_TABLE (table1), statusbar, 0, 1, 3, 4,
					(GtkAttachOptions)(GTK_EXPAND | GTK_FILL),  
					(GtkAttachOptions)0, 0, 0);
	gtk_widget_show(statusbar);

	hbox1 = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (hbox1);
	gtk_table_attach (GTK_TABLE (table1), hbox1, 0, 1, 0, 1,
					(GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
					(GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);

	vbox8 = gtk_vbox_new (FALSE, 0);
	gtk_widget_show (vbox8);

	hpaned = gtk_hpaned_new();
	gtk_widget_show (hpaned);

	gtk_paned_pack1(GTK_PANED(hpaned), vbox8, TRUE, FALSE);
	gtk_box_pack_start (GTK_BOX (hbox1), hpaned, TRUE, TRUE, 0);

	label8 = gtk_label_new ("NOTEBOOK");
	gtk_widget_show (label8);
	gtk_box_pack_start (GTK_BOX (vbox8), label8, FALSE, FALSE, 0);

	scrolledwindow5 = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_show (scrolledwindow5);
	gtk_box_pack_start (GTK_BOX (vbox8), scrolledwindow5, TRUE, TRUE, 0);


	viewport1 = gtk_viewport_new (NULL, NULL);
	gtk_widget_show (viewport1);
	gtk_container_add (GTK_CONTAINER (scrolledwindow5), viewport1);


	table3 = gtk_table_new (PAGE_ROWS + 1, NOTEBOOK_COLUMNS + 1, FALSE);
	gtk_widget_show (table3);
	gtk_container_add (GTK_CONTAINER (viewport1), table3);

	for(i = 1; i <= NOTEBOOK_COLUMNS; ++i)
		for(j = 1; j <= PAGE_ROWS; ++j) {
			notebook[i][j] = gtk_entry_new ();
			gtk_editable_set_editable((GtkEditable *)notebook[i][j], 0);
			gtk_entry_set_alignment(GTK_ENTRY(notebook[i][j]), 0.5);
			gtk_widget_show (notebook[i][j]);
			gtk_table_attach (GTK_TABLE (table3), notebook[i][j], i, i+1, j, j+1,
						   (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
						   (GtkAttachOptions) (0), 0, 0);
			gtk_widget_set_size_request(notebook[i][j], 100, -1);
		}

	for(j = 1; j <= PAGE_ROWS; ++j) {
		sprintf(temp, " (%d, %d) ", 0, j - 1);
		notebook[0][j] = gtk_label_new (temp);
		gtk_widget_show (notebook[0][j]);
		gtk_table_attach (GTK_TABLE (table3), notebook[0][j], 0, 1, j, j+1,
	    				 (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
						 (GtkAttachOptions) (0), 0, 0);
	}

	for(i = 0; i <= NOTEBOOK_COLUMNS; ++i) {
		switch(i) {
			case 0:	sprintf(temp, "(page, row)");	break;
			case 1:	sprintf(temp, "NAME");			break;
			case 2:	sprintf(temp, "VALUE");			break;
			case 3:	sprintf(temp, "TYPE");			break;
			case 4:	sprintf(temp, "STORAGE TYPE");	break;
			default: assert(false);
		}

		notebook[i][0] = gtk_label_new (temp);
		gtk_widget_show (notebook[i][0]);
		gtk_table_attach (GTK_TABLE (table3), notebook[i][0], i, i+1, 0, 1,
	    				 (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
						 (GtkAttachOptions) (0), 0, 0);
	}

	vbox9 = gtk_vbox_new (FALSE, 0);
	gtk_widget_show (vbox9);
	gtk_widget_set_size_request(vbox9, -1, 130);
	gtk_paned_pack2(GTK_PANED(hpaned), vbox9, FALSE, FALSE);

	button_previous = gtk_button_new_with_mnemonic ("PREVIOUS PAGE");
	gtk_widget_show (button_previous);
	gtk_box_pack_start (GTK_BOX (vbox9), button_previous, FALSE, FALSE, 0);

	button_next = gtk_button_new_with_mnemonic ("NEXT PAGE");
	gtk_widget_show (button_next);
	gtk_box_pack_start (GTK_BOX (vbox9), button_next, FALSE, FALSE, 0);


	frame = gtk_frame_new ("Select Page");
	gtk_widget_show(frame);
	gtk_box_pack_start (GTK_BOX (vbox9), frame, FALSE, FALSE, 30);

	box = gtk_vbox_new (FALSE, 0);
	gtk_widget_show(box);
    gtk_container_set_border_width (GTK_CONTAINER (box), 5);
    gtk_container_add (GTK_CONTAINER (frame), box);

	button_goto = gtk_button_new_with_mnemonic ("GOTO SELECTED PAGE");
	gtk_widget_show (button_goto);
	gtk_container_add (GTK_CONTAINER (box), button_goto);

	combo = gtk_combo_box_entry_new_text ();
	gtk_widget_set_size_request(combo, 50, -1);
	gtk_widget_show(combo);
	gtk_combo_box_append_text (GTK_COMBO_BOX (combo), "0");
	gtk_combo_box_append_text (GTK_COMBO_BOX (combo), "150");
	gtk_container_add (GTK_CONTAINER (box), combo);


	button_step = gtk_button_new_with_mnemonic ("STEP");
	gtk_widget_show (button_step);
	gtk_box_pack_end (GTK_BOX (vbox9), button_step, FALSE, FALSE, 0);

	hbox2 = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (hbox2);
	gtk_table_attach (GTK_TABLE (table1), hbox2, 0, 1, 1, 2,
				   (GtkAttachOptions) (GTK_FILL),
				   (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);

	vbox6 = gtk_vbox_new (FALSE, 0);
	gtk_widget_show (vbox6);
	gtk_box_pack_start (GTK_BOX (hbox2), vbox6, TRUE, TRUE, 0);

	label5 = gtk_label_new ("PROGRAM TRACE");
	gtk_widget_show (label5);
	gtk_box_pack_start (GTK_BOX (vbox6), label5, FALSE, FALSE, 0);

	scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_show (scrolledwindow1);
	gtk_box_pack_start (GTK_BOX (vbox6), scrolledwindow1, TRUE, TRUE, 0);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow1), GTK_SHADOW_IN);

	TraceText = gtk_text_view_new ();
	gtk_text_view_set_editable((GtkTextView *)TraceText, FALSE);
	gtk_widget_show (TraceText);
	gtk_container_add (GTK_CONTAINER (scrolledwindow1), TraceText);

	vbox1 = gtk_vbox_new (FALSE, 0);
	gtk_widget_show (vbox1);
	gtk_box_pack_start (GTK_BOX (hbox2), vbox1, TRUE, TRUE, 0);

	label6 = gtk_label_new ("INPUT");
	gtk_widget_show (label6);
	gtk_box_pack_start (GTK_BOX (vbox1), label6, FALSE, FALSE, 0);

	InputEntry = gtk_entry_new ();
	gtk_widget_show (InputEntry);
	gtk_box_pack_start (GTK_BOX (vbox1), InputEntry, FALSE, FALSE, 0);

	label7 = gtk_label_new ("OUTPUT");
	gtk_widget_show (label7);
	gtk_box_pack_start (GTK_BOX (vbox1), label7, FALSE, FALSE, 0);

	scrolledwindow2 = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_show (scrolledwindow2);
	gtk_box_pack_start (GTK_BOX (vbox1), scrolledwindow2, TRUE, TRUE, 0);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow2), GTK_SHADOW_IN);

	OutputText = gtk_text_view_new ();
	gtk_text_view_set_editable((GtkTextView *)OutputText, FALSE);
	gtk_widget_show (OutputText);
	gtk_container_add (GTK_CONTAINER (scrolledwindow2), OutputText);

	g_signal_connect ((gpointer) button_goto, "clicked",
					G_CALLBACK (OnGotoClicked), combo);
	g_signal_connect ((gpointer) button_previous, "clicked",
					G_CALLBACK (OnPreviousClicked), NULL);
	g_signal_connect ((gpointer) button_next, "clicked",
					G_CALLBACK (OnNextClicked), NULL);
	g_signal_connect ((gpointer) button_step, "clicked",
					G_CALLBACK (OnStepClicked), NULL);
	g_signal_connect ((gpointer) InputEntry, "activate",
					G_CALLBACK (OnInputActivate), NULL);

	return window1;
}

#endif	// HAS_VISUALIZER && GRAPHICAL_VISUALIZER
