// GraphicalVisualiser.cpp
// A graphical visualiser for the FLIP language.
// A. Savidis, October, 2005.
// Deep re-factoring of first version from Y. Lilis.
// Project: FLIP Execution Visualizer, Savidis, 2005.
//

#if	defined(HAS_VISUALIZER) && defined(GRAPHICAL_VISUALIZER)

#include "GraphicalVisualiser.h"
#include "GraphicalUILinkage.h"
#include "NotebookMirror.h"

#include <stdio.h>
#include <stdarg.h>
#include <assert.h>

VISUALISER_LINKAGE(GraphicalVisualiser)

////////////////////////////////////////////////////////////////

const char* GraphicalVisualiser::Input (const char* buff, unsigned len, FILE* fp) {

	// We ingore fp.
	//
	GRAPHICALUI::InputWait(const_cast<char*>(buff), len);
	return buff;
}

// Only program output.
//
int GraphicalVisualiser::Output (const char* format, ...) {

	va_list args;
	va_start(args, format);

	int result;
	static char buffer[1024];
	result = vsprintf(buffer, format, args);

	GRAPHICALUI::Output(buffer);

	va_end(args);
	return result;
}

////////////////////////////////////////////////////////////////

void GraphicalVisualiser::PreInitialiser (void) {
	GRAPHICALUI::Create(GetARGC(), GetARGV());
	NotebookMirror::SingletonCreate();
}

void GraphicalVisualiser::PostCleaner (void) {
	GRAPHICALUI::Destroy();
	NotebookMirror::SingletonDestroy();
}

void GraphicalVisualiser::CleanUp (void) {

	assert(!cleanedUp && visualiser);

	cleanedUp = true;
	delete ((GraphicalVisualiser*) visualiser);	// Derived class specific.
	visualiser = (VisualiserSuper*) 0;

	PostCleaner();
	exit(0);								// Clean-up exits immediately!
}

////////////////////////////////////////////////////////////////
// On program start, we explicitly wait upon start / end / stop.
//
void GraphicalVisualiser::OnProgramStart (void) {
	Output("Servant starts running the program.\n");

	GRAPHICALUI::Started();
	GetVisualiser()->StepWait();
}

void GraphicalVisualiser::OnProgramEnd (void) {
	Output("Servant finished the program normally");

	GRAPHICALUI::Finished();

	GetVisualiser()->StepWait();
	CleanUp();
}

void GraphicalVisualiser::OnProgramStop (unsigned line) {

	GRAPHICALUI::Trace(
		MakeString(
			"Servant stopped execution at line %d", line
		).c_str()
	);

	GRAPHICALUI::Finished();

	GetVisualiser()->StepWait();
	CleanUp();
}

void GraphicalVisualiser::OnProgramError (const char* s) {
	GRAPHICALUI::Trace(
		MakeString(
			"Error: %s\nServant has interrupted program execution", s
		).c_str()
	);

	GRAPHICALUI::Finished();

	GetVisualiser()->StepWait();
	CleanUp();
}

////////////////////////////////////////////////////////////////

GraphicalVisualiser::GraphicalVisualiser (void) {

	PreInitialiser();

	eventMutex = GRAPHICALUI::CreateMutex();

	ExecutionEvents::SetInputFunction(Input);
	ExecutionEvents::SetOutputFunction(Output);
	ExecutionEvents::SetCleanUpFunction(CleanUp);

	ExecutionEvents::SetOnProgramStart(OnProgramStart);
	ExecutionEvents::SetOnProgramEnd(OnProgramEnd);
	ExecutionEvents::SetOnProgramStop(OnProgramStop);
	ExecutionEvents::SetOnProgramError(OnProgramError);

	GRAPHICALUI::Trace(
		MakeString(
			"Servant starts running the program.\n"
			"Graphical execution tracer activated"
		).c_str()
	);
}

GraphicalVisualiser::~GraphicalVisualiser() {
	GRAPHICALUI::DestroyMutex(eventMutex );
}

////////////////////////////////////////////////////////////////

bool GraphicalVisualiser::IsNotifyingEvent (const std::string& name) {
	GRAPHICALUI::GrantMutex(eventMutex);
	bool result = VisualiserSuper::IsNotifyingEvent(name);
	GRAPHICALUI::ReleaseMutex(eventMutex);
	return result;
}

bool EVENTAPI::IsNotifyingEvent (const std::string& name)
	{ return VisualiserSuper::GetVisualiser()->IsNotifyingEvent(name); }

//****************************

void GraphicalVisualiser::SetNotifyingEvent (const std::string& name, bool enabled) {
	GRAPHICALUI::GrantMutex(eventMutex);
	VisualiserSuper::SetNotifyingEvent(name, enabled);
	GRAPHICALUI::ReleaseMutex(eventMutex);
}

void EVENTAPI::SetNotifyingEvent (const std::string& name, bool enabled)
	{ VisualiserSuper::GetVisualiser()->SetNotifyingEvent(name, enabled); }

//****************************

bool GraphicalVisualiser::IsSteppingEvent (const std::string& name) {
	GRAPHICALUI::GrantMutex(eventMutex);
	bool result = VisualiserSuper::IsSteppingEvent(name);
	GRAPHICALUI::ReleaseMutex(eventMutex);
	return result;
}

bool EVENTAPI::IsSteppingEvent (const std::string& name)
	{ return VisualiserSuper::GetVisualiser()->IsSteppingEvent(name); }

//****************************

void GraphicalVisualiser::SetSteppingEvent (const std::string& name, bool enabled) {
	GRAPHICALUI::GrantMutex(eventMutex);
	VisualiserSuper::SetSteppingEvent(name, enabled);
	GRAPHICALUI::ReleaseMutex(eventMutex);
}

void EVENTAPI::SetSteppingEvent (const std::string& name, bool enabled)
	{ VisualiserSuper::GetVisualiser()->SetSteppingEvent(name, enabled); }

////////////////////////////////////////////////////////////////

void GraphicalVisualiser::Message (const char* str) const
	{ GRAPHICALUI::Trace(str); }

//****************************

void GraphicalVisualiser::StepWait (void) const 
	{ GRAPHICALUI::StepWait(); }

//****************************

void GraphicalVisualiser::CellRefresh (
		const char*		name,	
		const char*		content,
		const char*		type, 
		const char*		storage, 
		const char*		operation, 
		unsigned		page, 
		unsigned		row
	)  {

	if (!name)
		name = MemGetObjectName(page, row);

	if (!storage)
		storage = MemGetObjectStorageString(page, row);

	NotebookMirror::GetSingleton()->GrantCellAccess();
	NotebookMirror::GetSingleton()->SetCellContent(page, row, name, content, type, storage);
	NotebookMirror::GetSingleton()->ReleaseCellAccess();

	Message(
		MakeString(
			"%s: \"%s\":%s = %s:%s\n\t at position (%d, %d)", 
			operation,
			name,
			storage,
			content,
			type,
			page,
			row
		).c_str()
	);

	// Now ask to refresh content.
	//
	GRAPHICALUI::Refresh(page, row);
}

//****************************

void GraphicalVisualiser::CellErase (
		unsigned	page, 
		unsigned	row
	)  {

	NotebookMirror::GetSingleton()->GrantCellAccess();
	NotebookMirror::GetSingleton()->EraseCell(page, row);
	NotebookMirror::GetSingleton()->ReleaseCellAccess();

	Message(
		MakeString(
			"Freeing: \"%s\":%s:%s\n\t from position (%d, %d)", 
			MemGetObjectName(page, row),
			MemGetObjectStorageString(page, row), 
			MemGetObjectTypeString(page, row),
			page,
			row
		).c_str()
	);

	// Again, ask to refresh content.
	//
	GRAPHICALUI::Refresh(page, row);
}

////////////////////////////////////////////////////////////////

#endif	// HAS_VISUALIZER && GTK_VISUALIZER